# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
   Change directory to provide relative paths for doctests
   >>> import os
   >>> filepath = os.path.dirname( os.path.realpath( __file__ ) )
   >>> datadir = os.path.realpath(os.path.join(filepath, '../testing/data'))
   >>> os.chdir(datadir)
"""

from __future__ import print_function
from __future__ import division

import os
import warnings

from nipype.interfaces.base import (
    TraitedSpec,
    CommandLineInputSpec,
    CommandLine,
    File,
    isdefined,
    traits,
)

warn = warnings.warn

pvc_methods = ['GTM',
               'IY',
               'IY+RL',
               'IY+VC',
               'LABBE',
               'LABBE+MTC',
               'LABBE+MTC+RL',
               'LABBE+MTC+VC',
               'LABBE+RBV',
               'LABBE+RBV+RL',
               'LABBE+RBV+VC',
               'MG',
               'MG+RL',
               'MG+VC',
               'MTC',
               'MTC+RL',
               'MTC+VC',
               'RBV',
               'RBV+RL',
               'RBV+VC',
               'RL',
               'VC']


class PETPVCInputSpec(CommandLineInputSpec):
    in_file   = File(desc="PET image file", exists=True, mandatory=True, argstr="-i %s")
    out_file  = File(desc="Output file", genfile=True, hash_files=False, argstr="-o %s")
    mask_file = File(desc="Mask image file", exists=True, mandatory=True, argstr="-m %s")
    pvc       = traits.Enum(pvc_methods, desc="Desired PVC method", mandatory=True, argstr="-p %s")
    fwhm_x    = traits.Float(desc="The full-width at half maximum in mm along x-axis", mandatory=True, argstr="-x %.4f")
    fwhm_y    = traits.Float(desc="The full-width at half maximum in mm along y-axis", mandatory=True, argstr="-y %.4f")
    fwhm_z    = traits.Float(desc="The full-width at half maximum in mm along z-axis", mandatory=True, argstr="-z %.4f")
    debug     = traits.Bool (desc="Prints debug information", usedefault=True, default_value=False, argstr="-d")
    n_iter    = traits.Int  (desc="Number of iterations", default_value=10, argstr="-n %d")
    n_deconv  = traits.Int  (desc="Number of deconvolution iterations", default_value=10, argstr="-k %d")
    alpha     = traits.Float(desc="Alpha value", default_value=1.5, argstr="-a %.4f")
    stop_crit = traits.Float(desc="Stopping criterion", default_value=0.01, argstr="-a %.4f")


class PETPVCOutputSpec(TraitedSpec):
    out_file = File(desc = "Output file")


class PETPVC(CommandLine):
    """ Use PETPVC for partial volume correction of PET images.

    PETPVC is a software from the Nuclear Medicine Department
    of the UCL University Hospital, London, UK.

    Its source code is here: https://github.com/UCL/PETPVC

    The methods that it implement are explained here:
    K. Erlandsson, I. Buvat, P. H. Pretorius, B. A. Thomas, and B. F. Hutton,
    "A review of partial volume correction techniques for emission tomography
    and their applications in neurology, cardiology and oncology," Phys. Med.
    Biol., vol. 57, no. 21, p. R119, 2012.

    There is a publication waiting to be accepted for this software tool.

    Its command line help shows this:

       -i --input < filename >
          = PET image file
       -o --output < filename >
          = Output file
       [ -m --mask < filename > ]
          = Mask image file
       -p --pvc < keyword >
          = Desired PVC method
       -x < X >
          = The full-width at half maximum in mm along x-axis
       -y < Y >
          = The full-width at half maximum in mm along y-axis
       -z < Z >
          = The full-width at half maximum in mm along z-axis
       [ -d --debug ]
          = Prints debug information
       [ -n --iter [ Val ] ]
          = Number of iterations
            With: Val (Default = 10)
       [ -k [ Val ] ]
          = Number of deconvolution iterations
            With: Val (Default = 10)
       [ -a --alpha [ aval ] ]
          = Alpha value
            With: aval (Default = 1.5)
       [ -s --stop [ stopval ] ]
          = Stopping criterion
            With: stopval (Default = 0.01)

    Technique - keyword
    -------------------
    - Geometric transfer matrix - "GTM"
    - Labbe approach - "LABBE"
    - Richardson-Lucy - "RL"
    - Van-Cittert - "VC"
    - Region-based voxel-wise correction - "RBV"
    - RBV with Labbe - "LABBE+RBV"
    - RBV with Van-Cittert - "RBV+VC"
    - RBV with Richardson-Lucy - "RBV+RL"
    - RBV with Labbe and Van-Cittert - "LABBE+RBV+VC"
    - RBV with Labbe and Richardson-Lucy- "LABBE+RBV+RL"
    - Multi-target correction - "MTC"
    - MTC with Labbe - "LABBE+MTC"
    - MTC with Van-Cittert - "MTC+VC"
    - MTC with Richardson-Lucy - "MTC+RL"
    - MTC with Labbe and Van-Cittert - "LABBE+MTC+VC"
    - MTC with Labbe and Richardson-Lucy- "LABBE+MTC+RL"
    - Iterative Yang - "IY"
    - Iterative Yang with Van-Cittert - "IY+VC"
    - Iterative Yang with Richardson-Lucy - "IY+RL"
    - Muller Gartner - "MG"
    - Muller Gartner with Van-Cittert - "MG+VC"
    - Muller Gartner with Richardson-Lucy - "MG+RL"

    Examples
    --------
    >>> from ..testing import example_data
    >>> #TODO get data for PETPVC
    >>> pvc = PETPVC()
    >>> pvc.inputs.in_file   = 'pet.nii.gz'
    >>> pvc.inputs.mask_file = 'tissues.nii.gz'
    >>> pvc.inputs.out_file  = 'pet_pvc_rbv.nii.gz'
    >>> pvc.inputs.pvc = 'RBV'
    >>> pvc.inputs.fwhm_x = 2.0
    >>> pvc.inputs.fwhm_y = 2.0
    >>> pvc.inputs.fwhm_z = 2.0
    >>> outs = pvc.run() #doctest: +SKIP
    """
    input_spec = PETPVCInputSpec
    output_spec = PETPVCOutputSpec
    _cmd = 'petpvc'

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self.inputs.out_file
        if not isdefined(outputs['out_file']):
            method_name = self.inputs.pvc.lower()
            outputs['out_file'] = self._gen_fname(self.inputs.in_file,
                                                  suffix='_{}_pvc'.format(method_name))

        outputs['out_file'] = os.path.abspath(outputs['out_file'])
        return outputs

    def _gen_fname(self, basename, cwd=None, suffix=None, change_ext=True,
                   ext='.nii.gz'):
        """Generate a filename based on the given parameters.

        The filename will take the form: cwd/basename<suffix><ext>.
        If change_ext is True, it will use the extentions specified in
        <instance>intputs.output_type.

        Parameters
        ----------
        basename : str
            Filename to base the new filename on.
        cwd : str
            Path to prefix to the new filename. (default is os.getcwd())
        suffix : str
            Suffix to add to the `basename`.  (defaults is '' )
        change_ext : bool
            Flag to change the filename extension to the given `ext`.
            (Default is False)

        Returns
        -------
        fname : str
            New filename based on given parameters.

        """
        from nipype.utils.filemanip import fname_presuffix

        if basename == '':
            msg = 'Unable to generate filename for command %s. ' % self.cmd
            msg += 'basename is not set!'
            raise ValueError(msg)
        if cwd is None:
            cwd = os.getcwd()
        if change_ext:
            if suffix:
                suffix = ''.join((suffix, ext))
            else:
                suffix = ext
        if suffix is None:
            suffix = ''
        fname = fname_presuffix(basename, suffix=suffix,
                                use_ext=False, newpath=cwd)
        return fname

    def _gen_filename(self, name):
        if name == 'out_file':
            return self._list_outputs()['out_file']
        return None



# -*- coding: utf-8 -*-
"""
Nipype workflows to process anatomical MRI.
"""
import json
from collections import OrderedDict

from nipype.utils.filemanip import filename_to_list
from nipype.interfaces.nipy.preprocess import Trim
import pandas as pd
from pydicom import read_file
from hansel import Crumb


def get_info(dicom_files):
    """Given a Siemens dicom file return metadata

    Returns
    -------
    RepetitionTime
    Slice Acquisition Times
    Spacing between slices
    """
    try:
        from dcmstack.extract import default_extractor
    except:
        raise ImportError("To use this function install dcmstack: pip install "
                          "git+https://github.com/moloney/dcmstack.git@c12d27d2c802d75a33ad70110124500a83e851ee#egg=dcmstack")

    meta = default_extractor(read_file(filename_to_list(dicom_files)[0],
                                       stop_before_pixels=True,
                                       force=True))
    return (meta['RepetitionTime'] / 1000., meta['CsaImage.MosaicRefAcqTimes'],
            meta['SpacingBetweenSlices'])


def trim(begin_idx=6):
    """Simple interface to trim a few volumes from a 4d fmri nifti file.

    :param begin_idx: int
        Remove first `begin_idx` volumes.

    Notes
    -----
    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.nipy.preprocess.html#trim
    """
    trim = Trim()
    #trim.inputs.in_file = 'functional.nii'
    trim.inputs.begin_index = begin_idx
    return trim


def motion_stats_sheet(motion_file_cr, crumb_fields):
    """ Return a pandas.DataFrame with some of the motion statistics obtained from the
    `statistics_files` output of the nipype.RapidArt found in the hansel.Crumb `motion_file_cr`.

    Parameters
    ----------
    motion_file_cr: str

    crumb_fields: list of str

    Returns
    -------
    df: pandas.DataFrame

    Examples
    --------
    >>> motion_stats_sheet(motion_file_cr="/home/hansel/data/thomas/out/{group}/{patient_id}/{session}/rest/artifact_stats/motion_stats.json", \
    >>>                    crumb_fields=['group', 'patient_id', 'session'])
    """
    def get_motion_record(mtn_file_cr, crumb_fields):
        """ Return an OrderedDict of the information found in the `mtn_file_cr` and also
        `crumb_fields` Crumb argument values."""
        stats = json.load(open(str(mtn_file_cr)))

        outliers = stats[1]
        motion_norm = stats[3]['motion_norm']

        #outliers_hdr = list(outliers.keys())
        motion_hdr   = ['{}_motion_norm'.format(k) for k in motion_norm.keys()]

        mtn_record = OrderedDict()
        for fn in crumb_fields:
            mtn_record[fn] = mtn_file_cr[fn][0]

        mtn_record.update(outliers)

        for hdr, fn in zip(motion_hdr, motion_norm):
            mtn_record[hdr] = motion_norm[fn]

        return mtn_record

    # process the input
    motion_file_cr = Crumb(motion_file_cr)
    crumb_fields   = [crf.strip() for crf in crumb_fields[1:-1].replace("'", "").split(',')]

    # create the motion records
    motionstats = [get_motion_record(stats_file, crumb_fields) for stats_file in motion_file_cr.ls()]

    # create a pandas Dataframe out of it
    df = pd.DataFrame.from_records(motionstats, columns=motionstats[0].keys())

    # return the dataframe
    return df



# -*- coding: utf-8 -*-
"""
Nipype workflow for cortical thickness measures using ANTs.
"""
import os.path as op

from   nipype.interfaces.ants import CorticalThickness

from   ..config  import setup_node
from   .._utils  import format_pair_list
from   ..utils   import (remove_ext,
                         extend_trait_list,
                         get_input_node,
                         get_datasink,
                         get_input_file_name,
                         extension_duplicates)


def attach_ants_cortical_thickness(main_wf, wf_name='ants_cortical_thickness'):
    """ Attach the ANTs anatomical cortical thickness workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Not used, this is there only to keep a uniform function signature.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.anat: input node

    cortical_thickness.brain_template: traits.File
        Anatomical *intensity* template (possibly created using a population
        data set with buildtemplateparallel.sh in ANTs).
        This template is *not* skull-stripped.

    cortical_thickness.segmentation_priors: string
        Pattern for paths to the tissue segmentation priors in the space of the
        given template.

        Specified using c-style formatting, e.g. "labelsPriors%02d.nii.gz".
        We assume that the first four priors are ordered as follows
            1:  csf
            2:  cortical gm
            3:  wm
            4:  deep gm

    corthick_input.brain_probability_mask: traits.File
        Brain probability mask in template space.

    cortical_thickness.t1_registration_template:
        Anatomical *intensity* template(assumed to be skull-stripped). A
        commoncase would be where this would be the sametemplate as
        specified in the -e option whichis not skull stripped.

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # cortical thickness node
    cort_thick = setup_node(CorticalThickness(), name="cortical_thickness")

    # The base name of the 'anat' file for the substitutions
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))

    # dataSink output substitutions
    regexp_subst = [
                   ]
    regexp_subst = format_pair_list(regexp_subst, anat=anat_fbasename)

    # add nii.gz patterns
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    main_wf.connect([(in_files, cort_thick, [("anat", "anatomical_image")]),

                     # output
                     (cort_thick, datasink, [("BrainExtractionMask",                "anat.cortex.@brain_mask"),
                                             ("BrainSegmentation",                  "anat.cortex.@brain_segmentation"),
                                             ("BrainSegmentationN4",                "anat.cortex.@brain_n4"),
                                             ("BrainSegmentationPosteriors",        "anat.cortex.@brain_seg_posteriors"),
                                             ("BrainVolumes",                       "anat.cortex.@brain_volumes"),
                                             ("CorticalThickness",                  "anat.cortex.@cortical_thickness"),
                                             ("CorticalThicknessNormedToTemplate",  "anat.cortex.@normed_cortical_thickness"),
                                             ("SubjectToTemplate0GenericAffine",    "anat.cortex.@subj_to_template_affine"),
                                             ("SubjectToTemplate1Warp",             "anat.cortex.@subj_to_template_warp"),
                                             ("SubjectToTemplateLogJacobian",       "anat.cortex.@subj_to_template_logjac"),
                                             ("TemplateToSubject0Warp",             "anat.cortex.@template_to_subj_warp"),
                                             ("TemplateToSubject1GenericAffine",    "anat.cortex.@template_to_subj_affine"),
                                            ]),
                    ])

    return main_wf



# -*- coding: utf-8 -*-
"""
Nipype workflow to clean up resting-state functional MRI.
"""
import nipype.pipeline.engine as pe
import os.path as op
from   nipype.algorithms.misc import Gunzip
from   nipype.interfaces import fsl
from   nipype.interfaces.nipy.preprocess import Trim, ComputeMask
from   nipype.interfaces.utility import Function, Select, IdentityInterface
from   neuro_pypes.interfaces.nilearn import mean_img, smooth_img

from   .filter import bandpass_filter
from   .nuisance import rest_noise_filter_wf
from   .._utils import format_pair_list, flatten_list
from   ..config import setup_node, get_config_setting
from   ..preproc import (auto_spm_slicetime,
                         nipy_motion_correction,
                         spm_coregister,
                         )
from   ..utils  import (remove_ext,
                        extend_trait_list,
                        get_input_node,
                        get_interface_node,
                        get_datasink,
                        get_input_file_name,
                        extension_duplicates)


def fmri_cleanup_wf(wf_name="fmri_cleanup"):
    """ Run the resting-state fMRI pre-processing workflow against the rest files in `data_dir`.

    Tasks:
    - Trim first 6 volumes of the rs-fMRI file.
    - Slice Timing correction.
    - Motion and nuisance correction.
    - Calculate brain mask in fMRI space.
    - Bandpass frequency filtering for resting-state fMRI.
    - Smoothing.
    - Tissue maps co-registration to fMRI space.

    Parameters
    ----------
    wf_name: str

    Nipype Inputs
    -------------
    rest_input.in_file: traits.File
        The resting-state fMRI file.

    rest_input.anat: traits.File
        Path to the high-contrast anatomical image.

    rest_input.tissues: list of traits.File
        Paths to the tissue segmentations in anatomical space.
        Expected to have this order: GM, WM and CSF.

    rest_input.highpass_sigma:traits.Float
        Band pass timeseries filter higher bound in Hz.

    rest_input.lowpass_sigma: traits.Float
        Band pass timeseries filter lower bound in Hz.

    Nipype Outputs
    --------------
    rest_output.smooth: traits.File
        The isotropically smoothed time filtered nuisance corrected image.

    rest_output.nuis_corrected: traits.File
        The nuisance corrected fMRI file.

    rest_output.motion_params: traits.File
        The affine transformation file.

    rest_output.time_filtered: traits.File
        The bandpass time filtered fMRI file.

    rest_output.epi_brain_mask: traits.File
        An estimated brain mask from mean EPI volume.

    rest_output.tissues_brain_mask: traits.File
        A brain mask calculated from the addition of coregistered
        GM, WM and CSF segmentation volumes from the anatomical
        segmentation.

    rest_output.tissues: list of traits.File
        The tissues segmentation volume in fMRI space.
        Expected to have this order: GM, WM and CSF.

    rest_output.anat: traits.File
        The T1w image in fMRI space.

    rest_output.avg_epi: traits.File
        The average EPI image in fMRI space after slice-time and motion correction.

    rest_output.motion_regressors: traits.File

    rest_output.compcor_regressors: traits.File

    rest_output.art_displacement_files
        One image file containing the voxel-displacement timeseries.

    rest_output.art_intensity_files
        One file containing the global intensity values determined from the brainmask.

    rest_output.art_norm_files
        One file containing the composite norm.

    rest_output.art_outlier_files
         One file containing a list of 0-based indices corresponding to outlier volumes.

    rest_output.art_plot_files
        One image file containing the detected outliers.

    rest_output.art_statistic_files
        One file containing information about the different types of artifacts and if design info is provided then
        details of stimulus correlated motion and a listing or artifacts by event type.

    Returns
    -------
    wf: nipype Workflow
    """
    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_file",
                  "anat",
                  "atlas_anat",
                  "coreg_target",
                  "tissues",
                  "lowpass_freq",
                  "highpass_freq",]

    out_fields = ["motion_corrected",
                  "motion_params",
                  "tissues",
                  "anat",
                  "avg_epi",
                  "time_filtered",
                  "smooth",
                  "tsnr_file",
                  "epi_brain_mask",
                  "tissues_brain_mask",
                  "motion_regressors",
                  "compcor_regressors",
                  "gsr_regressors",
                  "nuis_corrected",
                  "art_displacement_files",
                  "art_intensity_files",
                  "art_norm_files",
                  "art_outlier_files",
                  "art_plot_files",
                  "art_statistic_files",]

    # input identities
    rest_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                            name="rest_input")

    # rs-fMRI preprocessing nodes
    trim    = setup_node(Trim(), name="trim")

    stc_wf  = auto_spm_slicetime()
    realign = setup_node(nipy_motion_correction(), name='realign')

    # average
    average = setup_node(Function(function=mean_img, input_names=["in_file"], output_names=["out_file"],
                                  imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                         name='average_epi')

    mean_gunzip = setup_node(Gunzip(), name="mean_gunzip")

    # co-registration nodes
    coreg     = setup_node(spm_coregister(cost_function="mi"), name="coreg_fmri")
    brain_sel = setup_node(Select(index=[0, 1, 2]),            name="brain_sel")

    # brain mask made with EPI
    epi_mask    = setup_node(ComputeMask(),         name='epi_mask')

    # brain mask made with the merge of the tissue segmentations
    tissue_mask = setup_node(fsl.MultiImageMaths(), name='tissue_mask')
    tissue_mask.inputs.op_string = "-add %s -add %s -abs -kernel gauss 4 -dilM -ero -kernel gauss 1 -dilM -bin"
    tissue_mask.inputs.out_file = "tissue_brain_mask.nii.gz"

    # select tissues
    gm_select    = setup_node(Select(index=[0]),    name="gm_sel")
    wmcsf_select = setup_node(Select(index=[1, 2]), name="wmcsf_sel")

    # noise filter
    noise_wf   = rest_noise_filter_wf()
    wm_select  = setup_node(Select(index=[1]), name="wm_sel")
    csf_select = setup_node(Select(index=[2]), name="csf_sel")

    # bandpass filtering
    bandpass = setup_node(Function(input_names=['files',
                                                'lowpass_freq',
                                                'highpass_freq',
                                                'tr'],
                                   output_names=['out_files'],
                                   function=bandpass_filter),
                          name='bandpass')

    # smooth
    smooth = setup_node(Function(function=smooth_img,
                                 input_names=["in_file", "fwhm"],
                                 output_names=["out_file"],
                                 imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                         name="smooth")
    smooth.inputs.fwhm = get_config_setting('fmri_smooth.fwhm', default=8)
    smooth.inputs.out_file = "smooth_std_{}.nii.gz".format(wf_name)

    # output identities
    rest_output = setup_node(IdentityInterface(fields=out_fields),
                             name="rest_output")

    # Connect the nodes
    wf.connect([
                # trim
                (rest_input,   trim,         [("in_file", "in_file")]),

                #slice time correction
                (trim,        stc_wf,        [("out_file", "stc_input.in_file")]),

                # motion correction
                (stc_wf,      realign,       [("stc_output.timecorrected_files", "in_file")]),

                # coregistration target
                (realign,     average,       [("out_file", "in_file")]),
                (average,     mean_gunzip,   [("out_file", "in_file")]),
                (mean_gunzip, coreg,         [("out_file", "target")]),

                # epi brain mask
                (average,     epi_mask,      [("out_file", "mean_volume")]),

                # coregistration
                (rest_input,  coreg,         [("anat",                "source")]),
                (rest_input,  brain_sel,     [("tissues",             "inlist")]),
                (brain_sel,   coreg,         [(("out", flatten_list), "apply_to_files")]),

                # tissue brain mask
                (coreg,        gm_select,    [("coregistered_files",  "inlist")]),
                (coreg,        wmcsf_select, [("coregistered_files",  "inlist")]),
                (gm_select,    tissue_mask,  [(("out", flatten_list), "in_file")]),
                (wmcsf_select, tissue_mask,  [(("out", flatten_list), "operand_files")]),

                # nuisance correction
                (coreg,         wm_select,  [("coregistered_files",   "inlist",)]),
                (coreg,         csf_select, [("coregistered_files",   "inlist",)]),
                (realign,       noise_wf,   [("out_file",             "rest_noise_input.in_file",)]),
                (tissue_mask,   noise_wf,   [("out_file",             "rest_noise_input.brain_mask")]),
                (wm_select,     noise_wf,   [(("out", flatten_list),  "rest_noise_input.wm_mask")]),
                (csf_select,    noise_wf,   [(("out", flatten_list),  "rest_noise_input.csf_mask")]),

                (realign,       noise_wf,   [("par_file",             "rest_noise_input.motion_params",)]),

                # temporal filtering
                (noise_wf,    bandpass,    [("rest_noise_output.nuis_corrected", "files")]),
                #(realign,     bandpass,    [("out_file", "files")]),
                (stc_wf,      bandpass,    [("stc_output.time_repetition", "tr")]),
                (rest_input,  bandpass,    [("lowpass_freq",               "lowpass_freq"),
                                            ("highpass_freq",              "highpass_freq"),
                                           ]),
                (bandpass,     smooth,     [("out_files", "in_file")]),

                # output
                (epi_mask,    rest_output, [("brain_mask", "epi_brain_mask")]),
                (tissue_mask, rest_output, [("out_file",   "tissues_brain_mask")]),
                (realign,     rest_output, [("out_file",   "motion_corrected"),
                                            ("par_file",   "motion_params"),
                                           ]),
                (coreg,       rest_output, [("coregistered_files",  "tissues"),
                                            ("coregistered_source", "anat"),
                                           ]),
                (noise_wf,    rest_output, [("rest_noise_output.motion_regressors",      "motion_regressors"),
                                            ("rest_noise_output.compcor_regressors",     "compcor_regressors"),
                                            ("rest_noise_output.gsr_regressors",         "gsr_regressors"),
                                            ("rest_noise_output.nuis_corrected",         "nuis_corrected"),
                                            ("rest_noise_output.tsnr_file",              "tsnr_file"),
                                            ("rest_noise_output.art_displacement_files", "art_displacement_files"),
                                            ("rest_noise_output.art_intensity_files",    "art_intensity_files"),
                                            ("rest_noise_output.art_norm_files",         "art_norm_files"),
                                            ("rest_noise_output.art_outlier_files",      "art_outlier_files"),
                                            ("rest_noise_output.art_plot_files",         "art_plot_files"),
                                            ("rest_noise_output.art_statistic_files",    "art_statistic_files"),
                                           ]),
                (average,     rest_output, [("out_file",  "avg_epi")]),
                (bandpass,    rest_output, [("out_files", "time_filtered")]),
                (smooth,      rest_output, [("out_file",  "smooth")]),
              ])

    return wf


def attach_fmri_cleanup_wf(main_wf, wf_name="fmri_cleanup"):
    """ Attach the resting-state MRI pre-processing workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    registration_wf_name: str
        Name of the registration workflow.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.anat: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    anat_output = get_interface_node(main_wf, "anat_output")

    # create the fMRI preprocessing pipelines
    cleanup_wf = fmri_cleanup_wf(wf_name)

    # dataSink output substitutions
    # The base name of the 'rest' file for the substitutions
    rest_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'rest')))
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))

    regexp_subst = [
                    (r"/rc1[\w]+_corrected\.nii$",                                 "/gm_{rest}.nii"),
                    (r"/rc2[\w]+_corrected\.nii$",                                 "/wm_{rest}.nii"),
                    (r"/rc3[\w]+_corrected\.nii$",                                 "/csf_{rest}.nii"),
                    (r"/rm[\w]+_corrected\.nii$",                                  "/{anat}_{rest}.nii"),
                    (r"/corr_stc{rest}_trim\.nii$",                                "/slice_time_corrected.nii"),
                    (r"/stc{rest}_trim\.nii\.par$",                                "/motion_parameters.txt"),
                    (r"/corr_stc{rest}_trim_filt\.nii$",                           "/time_filt.nii"),
                    (r"/corr_stc{rest}_trim_mean_mask\.\.nii$",                    "/epi_brain_mask_{rest}.nii"),
                    (r"/tissue_brain_mask\.nii$",                                  "/tissue_brain_mask_{rest}.nii"),
                    (r"/corr_stc{rest}_trim_mean\.nii$",                           "/avg_epi.nii"),

                    (r"/art\..*_outliers\.txt$",                                   "/artifact_outliers.txt"),
                    (r"/global_intensity\..*\.txt$",                               "/global_intensities.txt"),
                    (r"/norm\..*_outliers\.txt$",                                  "/motion_norms.txt"),
                    (r"/stats\..*\.txt$",                                          "/motion_stats.json"),
                    (r"/plot\..*\.png$",                                           "/artifact_plots.png"),

                    (r"/corr_stc{rest}_trim_filtermotart\.nii$",                   "/{rest}_motion_corrected.nii"),
                    (r"/corr_stc{rest}_trim_filtermotart[\w_]*_cleaned\.nii$",     "/{rest}_nuisance_corrected.nii"),
                    (r"/corr_stc{rest}_trim_filtermotart[\w_]*_gsr\.nii$",         "/{rest}_nuisance_corrected.nii"),
                    (r"/corr_stc{rest}_trim_filtermotart[\w_]*_bandpassed\.nii$",  "/{rest}_time_filtered.nii"),
                    (r"/corr_stc{rest}_trim_filtermotart[\w_]*_smooth\.nii$",      "/{rest}_smooth.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, rest=rest_fbasename, anat=anat_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # input and output anat workflow to main workflow connections
    main_wf.connect([(in_files,   cleanup_wf, [("rest", "rest_input.in_file")]),

                    # anat to fMRI registration inputs
                    (anat_output, cleanup_wf, [("tissues_native", "rest_input.tissues"),
                                               ("anat_biascorr",  "rest_input.anat"),
                                              ]),

                    # clean_up_wf to datasink
                    (cleanup_wf,  datasink,  [
                                                ("rest_output.epi_brain_mask",         "rest.@epi_brain_mask"),
                                                ("rest_output.tissues_brain_mask",     "rest.@tissues_brain_mask"),
                                                ("rest_output.tissues",                "rest.@tissues"),
                                                ("rest_output.anat",                   "rest.@anat"),
                                                ("rest_output.motion_regressors",      "rest.@motion_regressors"),
                                                ("rest_output.compcor_regressors",     "rest.@compcor_regressors"),
                                                ("rest_output.gsr_regressors",         "rest.@gsr_regressors"),
                                                ("rest_output.motion_params",          "rest.@motion_params"),
                                                ("rest_output.motion_corrected",       "rest.@motion_corrected"),
                                                ("rest_output.nuis_corrected",         "rest.@nuis_corrected"),
                                                ("rest_output.time_filtered",          "rest.@time_filtered"),
                                                ("rest_output.smooth",                 "rest.@smooth"),
                                                ("rest_output.avg_epi",                "rest.@avg_epi"),
                                                ("rest_output.tsnr_file",              "rest.@tsnr"),
                                                ("rest_output.art_displacement_files", "rest.artifact_stats.@displacement"),
                                                ("rest_output.art_intensity_files",    "rest.artifact_stats.@art_intensity"),
                                                ("rest_output.art_norm_files",         "rest.artifact_stats.@art_norm"),
                                                ("rest_output.art_outlier_files",      "rest.artifact_stats.@art_outlier"),
                                                ("rest_output.art_plot_files",         "rest.artifact_stats.@art_plot"),
                                                ("rest_output.art_statistic_files",    "rest.artifact_stats.@art_statistic"),
                                               ]),
                    ])

    return main_wf



# -*- coding: utf-8 -*-
"""
Helper functions for joining, merging, managing the workflow nodes.
"""

from nipype.interfaces.utility import Function, IdentityInterface
from nipype.interfaces.io import SelectFiles, DataSink, DataGrabber
from nipype.interfaces.base import traits, isdefined
import nipype.interfaces.fsl as fsl

from ..crumb  import DataCrumb


def get_trait_value(traitspec, value_name, default=None):
    """ Return the attribute `value_name` from traitspec if it is defined.
    If not will return the value of `default`.
    Parameters
    ----------
    traitspec: TraitedSpec

    value_name: str
        Name of the `traitspect` attribute.

    default: any
        A default value in case the attribute does not exist or is not defined.

    Returns
    -------
    trait_value: any
    """
    val = getattr(traitspec, value_name, default)
    return default if not isdefined(val) else val


def selectindex(files, idx, flatten=True):
    """ Select the items in the list `files` in the indexes given by the list of
    integers `idx`."""
    import numpy as np
    from nipype.utils.filemanip import filename_to_list, list_to_filename
    from neuro_pypes._utils import flatten_list

    if flatten:
        files = flatten_list(files)

    if isinstance(idx, list):
        return list_to_filename(np.array(filename_to_list(files))[idx].tolist())
    else:
        return files[idx]


def get_node(wf, node_types, name=''):
    """ Return the first node found in `wf` of any of the types in `node_types`.

    Parameters
    ----------
    wf: nipype Workflow

    node_types: sequence of node types

    Returns
    -------
    nodes: nipype Node
        The DataSink in `wf`.
    """
    node = None
    for ionode in node_types:
        node = find_wf_node(wf, ionode, name=name)
        if node is not None:
            return node

    if node is None:
        raise KeyError('Could not find a node of type {} in the worflow.'.format(node_types))


def get_datasink(wf, name=''):
    """ Return the DataSink node in `wf`

    Parameters
    ----------
    wf: nipype Workflow

    Returns
    -------
    nodes: nipype Node
        The DataSink in `wf`.
    """
    return get_node(wf, (DataSink, ), name=name)


def get_input_node(wf, name=''):
    """ Return the first node of type: (DataCrumb, SelectFiles, DataGrabber) in wf."""
    return get_node(wf, (DataCrumb, SelectFiles, DataGrabber), name=name)


def get_interface_node(wf, name):
    """ Return the first node found of type IdentityInterface in `wf` with name `name`."""
    return get_node(wf, (IdentityInterface, ), name=name)


def wf_nodes(wf, iface_type):
    """ Return the nodes found in the list of node names in `wf` that
    has an interface of type `iface_type`.

    Parameters
    ----------
    wf: nipype Workflow

    iface_type: nipype class

    Returns
    -------
    nodes: list of nipype Node
    """
    nodes = []
    for name in wf.list_node_names():
        if isinstance(getattr(wf.get_node(name), 'interface', None), iface_type):
            nodes.append(wf.get_node(name))
    return nodes


def find_wf_node(wf, iface_type, name=''):
    """ Return the first node found in the list of node names in `wf` that
    has an interface of type `iface_type`.

    Parameters
    ----------
    wf: nipype Workflow

    iface_type: nipype class

    Returns
    -------
    node: nipype Node
    """
    for node_nom in wf.list_node_names():
        if isinstance(getattr(wf.get_node(node_nom), 'interface', None), iface_type):
            if not name:
                return wf.get_node(node_nom)
            elif name in node_nom.split('.'):
                return wf.get_node(node_nom)
    return None


def extend_trait_list(trait_list, extension):
    """ Extend or initialize `trait_list` with `extension`.

    Parameters
    ----------
    trait_list: traits.List

    extension: list or tuple

    Returns
    -------
    trait_list: traits.List
        The extended list.
    """
    if isdefined(trait_list):
        if not isinstance(extension, (list, tuple)):
            raise ValueError('Expected `extension` to be list or tuple, got {}.'.format(type(extension)))

        trait_list.extend(extension)
    else:
        trait_list = extension

    return trait_list


def sum2args():
    """ Return a nipype function that sums up two args: `arg1` and `arg2` and
    leaves the result in `out`.

    Returns
    -------
    fi: nipype.interfaces.Function
    """
    func = 'def func(arg1, arg2): return arg1 + arg2'
    fi = Function(input_names=['arg1', 'arg2'], output_names=['out'])
    fi.inputs.function_str = func
    return fi


def joinstrings(n_args=2):
    """ Return a nipype function that joins up to `n_args` parts. and
    leaves the result in `out`.
    Parameters
    ----------
    n_args: int
        Number of `argX` parameters the function might have.
        Example: If `n_args` == 2, then the node will have `arg1` and `arg2` nodes.

    Returns
    -------
    fi: nipype.interfaces.Function
    """
    arg_names = ['arg{}'.format(n) for n in range(1, n_args+1)]

    func = '''def func({0}): import os; return os.path.join({0})'''.format(', '.join(arg_names))
    fi = Function(input_names=arg_names, output_names=['out'])
    fi.inputs.function_str = func
    return fi


def fsl_merge(in_files=traits.Undefined, dimension='t'):
    """ Merge the NifTI files in `in_files` in the given `dimension`.
    This uses `fslmerge`.

    Parameters
    ----------
    in_files: list of str.
        Paths to the files to merge.

    dimension: str
        Character indicating the merging dimension.
        Choices: 't', 'x', 'y', 'z'

    Returns
    -------
    merger: fsl.Merge
    """
    merger = fsl.Merge()
    merger.inputs.dimension = dimension
    merger.inputs.output_type = "NIFTI_GZ"
    merger.inputs.in_files = in_files

    return merger


def get_input_file_name(input_node, fname_key):
    """ Return the name of the file given by the node key `fname_key` in `input_node`.

    Parameters
    ----------
    input_node: nipype Node
        a node with a file input interface (SelectFiles, DataCrumb), for now.

    fname_key: str
        The key that is used to access the file path using `input_node`.
        Example: 'anat'

    Returns
    -------
    filename: str
        The path template of the input file `fname_key` path from the `input_node`.
    """
    if isinstance(input_node.interface, SelectFiles):
        try:
            fname = input_node.interface._templates[fname_key]
        except AttributeError:
            raise
        except KeyError:
            raise KeyError('Could not find file key {} in node {}({}).'.format(fname_key, input_node.name,
                                                                           type(input_node.interface)))
        else:
            return fname

    if isinstance(input_node.interface, DataCrumb):
        try:
            crumb_args = input_node.interface._templates[fname_key]
            incrumb    = input_node.interface._crumb.replace(**dict(crumb_args))
        except AttributeError:
            raise
        except KeyError:
            raise KeyError('Could not find file key {} in node {}({}).'.format(fname_key, input_node.name,
                                                                           type(input_node.interface)))
        else:
            return incrumb.path

    else:
        raise NotImplementedError('`get_input_file_name` has not been implemented for nodes'
                                  ' of type {}.'.format(type(input_node.interface)))


# -*- coding: utf-8 -*-
"""
Helper functions to check for value configurations in the system and the external tools.
"""
import os
from os import path as op

from nipype.interfaces import spm

from ..config import get_config_setting


def spm_tpm_priors_path(spm_dir=None):
    """ Return the path to the TPM.nii file from SPM.

    Parameters
    ----------
    spm_dir: str
        Path to SPM.
        If `None` will try nipype to guess it. #TOBEDONE

    Returns
    -------
    tpm_path: str
        Path to the TPM.nii file.

    Raises
    ------
    NotADirectoryError
        If the SPM path is not found.

    FileNotFoundError
        If `template` is `None` and can't find the TPM.nii file from SPM.
    """
    matlab_cmd = os.environ.get('SPMMCRCMD')
    spm_version = spm.Info.version(matlab_cmd=matlab_cmd,
                                   use_mcr='FORCE_SPMMCR' in os.environ)

    if spm_version is None:
        raise RuntimeError("Nipype could not find a valid Matlab or SPM configuration.")

    if spm_dir is None:
        spm_dir = spm_version.get('path', None)

    if spm_dir is None:
        spm_dir = op.expanduser(get_config_setting('spm_dir'))

    if not spm_dir:
        raise NotADirectoryError('Could not find SPM path.')

    if not op.exists(spm_dir):
        raise NotADirectoryError('The specified SPM path ({}) does not exist.'.format(spm_dir))

    tpm_path = op.join(spm_dir, 'tpm', 'TPM.nii')
    if not op.exists(tpm_path):
        raise FileNotFoundError('Could not find TPM.nii file from SPM.')

    return tpm_path


def get_fsl_extension(default='NIFTI_GZ'):
    """ Return the file extension that FSL will output.

    Parameters
    ----------
    default: str
        Default output type for FSL in FSL format.
        Choices: ANALYZE, NIFTI, NIFTI_PAIR,
                ANALYZE_GZ, NIFTI_GZ, NIFTI_PAIR_GZ

    Returns
    -------
    ext: str
        The file extension

    Note
    ----
    In the case of the pair files, this will only return
    the extension of the header file.

    #TODO: deal with pair file renaming with header modification.
    Check boyle for this.
    """
    outtype = os.environ.get('FSLOUTPUTTYPE', default)

    otype_ext = {'NIFTI':           '.nii',
                 'NIFTI_GZ':        '.nii.gz',
                 'ANALYZE':         '.hdr',
                 'NIFTI_PAIR':      '.hdr',
                 'NIFTI_PAIR_GZ':   '.hdr.gz',
                 'ANALYZE_PAIR_GZ': '.hdr.gz',
                }

    return otype_ext[outtype]



# -*- coding: utf-8 -*-
"""
Helper functions for motion correction of 4D images.
"""
from nipype import Workflow, MapNode
from nipype.interfaces.utility import IdentityInterface
from nipype.interfaces.base import traits
from nipype.interfaces.nipy import SpaceTimeRealigner
from nipype.interfaces.fsl import MCFLIRT, Split, Merge, ApplyWarp

from ..config import setup_node


def nipy_motion_correction(in_file=traits.Undefined):
    """
    Simultaneous motion and slice timing correction algorithm
    If slice_times is not specified, this algorithm performs spatial motion correction
    This interface wraps nipy’s SpaceTimeRealign algorithm [Roche2011] or
    simply the SpatialRealign algorithm when timing info is not provided.

    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.nipy.preprocess.html#spacetimerealigner

    Returns
    -------
    realign: nipype Interface

    Notes
    -----
    Roche A. A four-dimensional registration algorithm with application to joint correction of motion and slice
    timing in fMRI. IEEE Trans Med Imaging. 2011 Aug;30(8):1546-54. http://dx.doi.org/10.1109/TMI.2011.2131152
    """
    #Run spatial realignment only
    realigner = SpaceTimeRealigner()
    realigner.inputs.in_file = in_file

    #res = realigner.run()
    return realigner


def select_volume(filename, which):
    """Return the middle index of a file
    """
    from nibabel import load
    import numpy as np
    if which.lower() == 'first':
        idx = 0
    elif which.lower() == 'middle':
        idx = int(np.ceil(load(filename).shape[3] / 2))
    else:
        raise Exception('unknown value for volume selection : %s' % which)
    return idx


def fsl_motion_correction(name='realign'):
    """Realign a time series to the middle volume using spline interpolation

    Uses MCFLIRT to realign the time series and ApplyWarp to apply the rigid
    body transformations using spline interpolation (unknown order).

    Nipype Inputs
    -------------
    realign_input.func: exising file
        The path to the fMRI file.

    Nipype Outputs
    --------------
    realign_output.realigned_file: exising file
        The path to the realigned fMRI file.

    realign_output.realign_params: mat file
        .mat file with the affine transformation parameters.

    Example
    -------
    >>> wf = fsl_motion_correction()
    >>> wf.inputs.inputspec.func = 'f3.nii'
    >>> wf.run() # doctest: +SKIP
    """
    wf = Workflow(name=name)

    # input node
    input = setup_node(IdentityInterface(fields=['func']),
                       name='realign_input')
    realigner = setup_node(MCFLIRT(save_mats=True, stats_imgs=True),
                           name='mcflirt')
    splitter  = setup_node(Split(dimension='t'),
                           name='splitter')
    warper    = MapNode(ApplyWarp(interp='spline'),
                        iterfield=['in_file', 'premat'],
                        name='warper')

    joiner    = setup_node(Merge(dimension='t'),
                           name='joiner')

    # output node
    output = setup_node(IdentityInterface(fields=['realigned_file',
                                                  'realign_params',
                                                 ]),
                        name='realign_output')

    wf.connect([
                # input
                (input,     realigner, [("func",                             "in_file"),
                                        (("func", select_volume, 'middle'),  "ref_vol"),
                                       ]),
                # split
                (realigner,  splitter, [("out_file",      "in_file")]),
                (realigner,  warper,   [("mat_file",      "premat"),
                                        ("variance_img",  "ref_file"),
                                       ]),
                # warp
                (splitter,  warper,    [("out_files",      "in_file")]),
                (warper,    joiner,    [("mat_file",       "premat")]),
                # output
                (joiner,    output,    [("merged_file", "realigned_file")]),
                (realigner, output,    [("mat_file",    "realign_params")]),
               ])
    return wf



# -*- coding: utf-8 -*-
"""
Nipype workflows to process resting-state functional MRI.
"""
import os.path as op

import nipype.pipeline.engine    as pe
from   nipype.algorithms.misc    import Gunzip
from   nipype.interfaces.utility import Function, Merge, IdentityInterface
from   nipype.interfaces import spm, fsl

from   .._utils  import format_pair_list
from   ..config  import setup_node, get_config_setting, check_atlas_file
from   ..preproc import (spm_normalize,
                         get_bounding_box,
                         spm_tpm_priors_path,
                         spm_coregister,
                         spm_apply_deformations,
                         )
from   ..utils   import (remove_ext,
                         selectindex,
                         extend_trait_list,
                         get_input_node,
                         get_datasink,
                         get_input_file_name,
                         extension_duplicates)


def spm_warp_fmri_wf(wf_name="spm_warp_fmri", register_to_grptemplate=False):
    """ Run SPM to warp resting-state fMRI pre-processed data to MNI or a given 
    template.

    Tasks:
    - Warping the inputs to MNI or a template, if `do_group_template` is True

    Parameters
    ----------
    wf_name: str

    register_to_grptemplate: bool
        If True will expect the wfmri_input.epi_template input and use it as a group template
        for inter-subject registratio.

    Nipype Inputs
    -------------
    wfmri_input.in_file: traits.File
        The slice time and motion corrected fMRI file.

    wfmri_input.reference_file: traits.File
        The anatomical image in its native space
        for registration reference.

    wfmri_input.anat_fmri: traits.File
        The anatomical image in fMRI space.

    wfmri_input.anat_to_mni_warp: traits.File
        The warp field from the transformation of the
        anatomical image to the standard MNI space.

    wfmri_input.time_filtered: traits.File
        The bandpass time filtered fMRI file.

    wfmri_input.avg_epi: traits.File
        The average EPI from the fMRI file.

    wfmri_input.epi_template: traits.File
        Reference EPI template file for inter subject registration.
        If `do_group_template` is True you must specify this input.

    wfmri_input.brain_mask: traits.File
        Brain mask in fMRI space.

    wfmri_input.atlas_anat: traits.File
        Atlas in subject anatomical space.

    Nipype Outputs
    --------------
    wfmri_output.warped_fmri: traits.File
        The slice time, motion, and nuisance corrected fMRI
        file registered to the template.

    wfmri_output.wtime_filtered: traits.File
        The bandpass time filtered fMRI file
        registered to the template.

    wfmri_output.smooth: traits.File
        The smooth bandpass time filtered fMRI file
        registered to the template.

    wfmri_output.wavg_epi: traits.File
        The average EPI from the fMRI file
        registered to the template.

    wfmri_output.warp_field: traits.File
        The fMRI to template warp field.

    wfmri_output.coreg_avg_epi: traits.File
        The average EPI image in anatomical space.

        Only if registration.anat2fmri is false.

    wfmri_output.coreg_others: traits.File
        Other mid-preprocessing fmri images registered to
        anatomical space:

        - wfmri_input.in_file,

        - wfmri_input.brain_mask,

        - wfmri_input.time_filtered.

        Only if registration.anat2fmri is false

    wfmri_output.wbrain_mask: traits.File
        Brain mask in fMRI space warped to MNI.

    Returns
    -------
    wf: nipype Workflow
    """
    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_file",
                  "anat_fmri",
                  "anat_to_mni_warp",
                  "brain_mask",
                  "reference_file",
                  "time_filtered",
                  "avg_epi",]

    out_fields = ["warped_fmri",
                  "wtime_filtered",
                  "smooth",
                  "wavg_epi",
                  "wbrain_mask",
                  "warp_field",
                  "coreg_avg_epi",
                  "coreg_others"
                  ]

    if register_to_grptemplate:
        in_fields += ['epi_template']

    do_atlas, _ = check_atlas_file()
    if do_atlas:
        in_fields  += ["atlas_anat"]
        out_fields += ["atlas_fmri"]

    # input identities
    wfmri_input = setup_node(IdentityInterface(fields=in_fields, mandatory_inputs=True),
                             name="wfmri_input")

    # in file unzipper
    in_gunzip = pe.Node(Gunzip(), name="in_gunzip")

    # merge list for normalization input
    merge_list = pe.Node(Merge(2), name='merge_for_warp')
    gunzipper = pe.MapNode(Gunzip(), name="gunzip", iterfield=['in_file'])

    # the template bounding box
    tpm_bbox = setup_node(Function(function=get_bounding_box,
                                   input_names=["in_file"],
                                   output_names=["bbox"]),
                          name="tpm_bbox")

    # smooth the final result
    smooth = setup_node(fsl.IsotropicSmooth(fwhm=8, output_type='NIFTI'), name="smooth_fmri")

    # output identities
    rest_output = setup_node(IdentityInterface(fields=out_fields),
                             name="wfmri_output")


    # check how to perform the registration, to decide how to build the pipeline
    anat2fmri = get_config_setting('registration.anat2fmri', False)
    # register to group template
    if register_to_grptemplate:
        gunzip_template = pe.Node(Gunzip(), name="gunzip_template",)
        warp = setup_node(spm.Normalize(jobtype="estwrite", out_prefix="wgrptmpl_"),
                          name="fmri_grptemplate_warp",)
        warp_source_arg    = "source"
        warp_outsource_arg = "normalized_source"
        warp_field_arg     = "normalization_parameters"

    elif anat2fmri:
        # register to standard template
        warp = setup_node(spm_normalize(), name="fmri_warp")
        tpm_bbox.inputs.in_file = spm_tpm_priors_path()
        warp_source_arg    = "image_to_align"
        warp_outsource_arg = "normalized_image"
        warp_field_arg     = "deformation_field"

    else: # anat2fmri is False
        coreg       = setup_node(spm_coregister(cost_function="mi"), name="coreg_fmri")
        warp        = setup_node(spm_apply_deformations(), name="fmri_warp")
        coreg_files = pe.Node(Merge(3), name='merge_for_coreg')
        warp_files  = pe.Node(Merge(2), name='merge_for_warp')
        tpm_bbox.inputs.in_file = spm_tpm_priors_path()

    # make the connections
    if register_to_grptemplate:
        wf.connect([
                    # get template bounding box to apply to results
                    (wfmri_input, tpm_bbox, [("epi_template", "in_file")]),

                    # unzip and forward the template file
                    (wfmri_input,     gunzip_template, [("epi_template", "in_file")]),
                    (gunzip_template, warp,            [("out_file",     "template")]),

                    # get template bounding box to apply to results
                    (wfmri_input, tpm_bbox, [("epi_template", "in_file")]),
                    ])

    if anat2fmri or register_to_grptemplate:
        # prepare the inputs
        wf.connect([
                    # unzip the in_file input file
                    (wfmri_input, in_gunzip, [("avg_epi", "in_file")]),

                    # warp source file
                    (in_gunzip, warp, [("out_file", warp_source_arg)]),

                    # bounding box
                    (tpm_bbox,  warp, [("bbox", "write_bounding_box")]),

                    # merge the other input files into a list
                    (wfmri_input, merge_list, [("in_file",       "in1"),
                                               ("time_filtered", "in2"),
                                              ]),

                    # gunzip them for SPM
                    (merge_list, gunzipper, [("out", "in_file")]),

                    # apply to files
                    (gunzipper, warp,       [("out_file", "apply_to_files")]),

                    # outputs
                    (warp, rest_output,     [(warp_field_arg,     "warp_field"),
                                             (warp_outsource_arg, "wavg_epi"),
                                            ]),

                   ])

    else: # FMRI to ANAT
        wf.connect([
                    (wfmri_input, coreg,      [("reference_file", "target")]),

                    # unzip the in_file input file
                    (wfmri_input, in_gunzip,  [("avg_epi",  "in_file")]),
                    (in_gunzip,   coreg,      [("out_file", "source")]),

                    # merge the other input files into a list
                    (wfmri_input, coreg_files, [("in_file",       "in1"),
                                                ("time_filtered", "in2"),
                                                ("brain_mask",    "in3"),
                                               ]),

                    # gunzip them for SPM
                    (coreg_files, gunzipper, [("out", "in_file")]),

                    # coregister fmri to anat
                    (gunzipper,   coreg,     [("out_file", "apply_to_files")]),

                    # anat to mni warp field
                    (wfmri_input, warp, [("anat_to_mni_warp", "deformation_file")]),

                    # bounding box
                    (tpm_bbox,  warp, [("bbox", "write_bounding_box")]),

                    # apply to files
                    (coreg, warp_files, [("coregistered_source", "in1")]),
                    (coreg, warp_files, [("coregistered_files",  "in2")]),

                    (warp_files, warp, [("out", "apply_to_files")]),

                    # outputs
                    (warp, rest_output,  [("normalized_files",  "warped_files"),]),
                    (warp, rest_output,  [(("normalized_files", selectindex, 0), "wavg_epi"),]),

                    (coreg, rest_output, [("coregistered_source", "coreg_avg_epi")]),
                    (coreg, rest_output, [("coregistered_files",  "coreg_others")]),
                   ])

    # atlas file in fMRI space
    if anat2fmri:
        coreg_atlas = setup_node(spm_coregister(cost_function="mi"), name="coreg_atlas2fmri")

        # set the registration interpolation to nearest neighbour.
        coreg_atlas.inputs.write_interp = 0
        wf.connect([
                    (wfmri_input, coreg_atlas, [("reference_file", "source"),
                                                ("atlas_anat", "apply_to_files"),
                                               ]),
                    (in_gunzip,   coreg_atlas, [("out_file",           "target")]),
                    (coreg_atlas, rest_output, [("coregistered_files", "atlas_fmri")]),
                  ])

    # smooth and sink
    wf.connect([
                # smooth the final bandpassed image
                (warp,   smooth,      [(("normalized_files", selectindex, 1), "in_file")]),

                # output
                (smooth, rest_output, [("out_file", "smooth")]),
                (warp,   rest_output, [(("normalized_files", selectindex, 0), "warped_fmri"),
                                       (("normalized_files", selectindex, 1), "wtime_filtered"),
                                      ]),
               ])

    return wf


def attach_spm_warp_fmri_wf(main_wf, registration_wf_name="spm_warp_fmri", do_group_template=False):
    """ Attach the fMRI inter-subject spatial normalization workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    registration_wf_name: str
        Name of the registration workflow.

    do_group_template: bool
        If True will attach the group template creation and pre-processing pipeline.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.anat: input node

    datasink: nipype Node

    Workflow Dependencies
    ---------------------
    fmri_cleanup, the cleanup and preprocessing of the fMRI data

    spm_anat_preproc, for the anatomical to MNI space transformation

    spm_fmri_template, if do_group_template is True

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    anat_wf    = main_wf.get_node("spm_anat_preproc")
    cleanup_wf = main_wf.get_node("fmri_cleanup")

    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    if do_group_template:
        template_name = 'grptemplate'
    else:
        template_name = 'stdtemplate'

    warp_wf_name = "{}_{}".format(registration_wf_name, template_name)
    warp_fmri_wf = spm_warp_fmri_wf(warp_wf_name, register_to_grptemplate=do_group_template)

    # dataSink output substitutions
    # The base name of the 'rest' file for the substitutions
    rest_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'rest')))
    anat_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'anat')))

    regexp_subst = [
                    (r"/corr_stc{fmri}_trim_mean_sn.mat$", "/{fmri}_grptemplate_params.mat"),
                    (r"/y_corr_stc{fmri}_trim_mean\.nii$", "/{fmri}_to_mni_warpfield.nii"),
                    (r"/rcorr_stc{fmri}_trim_mean.nii$",   "/avg_epi_anat.nii"),

                    (r"/wgrptmpl_corr_stc{fmri}_trim_mean\.nii$",                          "/avg_epi_grptemplate.nii"),
                    (r"/wgrptmpl_corr_stc{fmri}_trim\.nii$",                               "/{fmri}_trimmed_grptemplate.nii"),
                    (r"/wgrptmpl_corr_stc{fmri}_trim_filtermotart[\w_]*_cleaned\.nii$",    "/{fmri}_nuisance_corrected_grptemplate.nii"),
                    (r"/wgrptmpl_corr_stc{fmri}_trim_filtermotart[\w_]*_gsr\.nii$",        "/{fmri}_nuisance_corrected_grptemplate.nii"),
                    (r"/wgrptmpl_corr_stc{fmri}_trim_filtermotart[\w_]*_bandpassed\.nii$", "/{fmri}_time_filtered_grptemplate.nii"),
                    (r"/wgrptmpl_corr_stc{fmri}_trim_filtermotart[\w_]*_smooth\.nii$",     "/{fmri}_smooth_grptemplate.nii"),

                    (r"/w[r]?corr_stc{fmri}_trim_mean\.nii$",                          "/avg_epi_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim\.nii$",                               "/{fmri}_trimmed_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim_filtermotart[\w_]*_cleaned\.nii$",    "/{fmri}_nuisance_corrected_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim_filtermotart[\w_]*_gsr\.nii$",        "/{fmri}_nuisance_corrected_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim_filtermotart[\w_]*_bandpassed\.nii$", "/{fmri}_time_filtered_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim_filtermotart[\w_]*_smooth\.nii$",     "/{fmri}_smooth_mni.nii"),

                    (r"/w[r]?corr_stc{fmri}_trim[\w_]*_smooth\.nii$",       "/{fmri}_nofilt_smooth_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim[\w_]*_cleaned\.nii$",      "/{fmri}_nofilt_nuisance_corrected_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim[\w_]*_gsr\.nii$",          "/{fmri}_nofilt_nuisance_corrected_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim[\w_]*_bandpassed\.nii$",   "/{fmri}_nofilt_time_filtered_mni.nii"),
                    (r"/w[r]?corr_stc{fmri}_trim[\w_]*_smooth\.nii$",       "/{fmri}_nofilt_smooth_mni.nii"),
                  ]
    regexp_subst = format_pair_list(regexp_subst, fmri=rest_fbasename, anat=anat_fbasename)

    # prepare substitution for atlas_file, if any
    do_atlas, atlas_file = check_atlas_file()
    if do_atlas:
        atlas_basename = remove_ext(op.basename(atlas_file))
        regexp_subst.extend([
                             (r"/[\w]*{atlas}.*\.nii$", "/{atlas}_{fmri}_space.nii"),
                            ])
        regexp_subst = format_pair_list(regexp_subst, atlas=atlas_basename,
                                                      fmri=rest_fbasename)

    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    # input and output anat workflow to main workflow connections
    main_wf.connect([# clean_up_wf to registration_wf
                     (cleanup_wf, warp_fmri_wf, [
                                                 ("rest_output.motion_corrected",   "wfmri_input.in_file"),
                                                 ("rest_output.anat",               "wfmri_input.anat_fmri"),
                                                 ("rest_output.time_filtered",      "wfmri_input.time_filtered"),
                                                 ("rest_output.avg_epi",            "wfmri_input.avg_epi"),
                                                 ("rest_output.tissues_brain_mask", "wfmri_input.brain_mask"),
                                                ]),
                     # output
                     (warp_fmri_wf,  datasink,  [
                                                 ("wfmri_output.warped_fmri",    "rest.{}.@warped_fmri".format(template_name)),
                                                 ("wfmri_output.wtime_filtered", "rest.{}.@time_filtered".format(template_name)),
                                                 ("wfmri_output.smooth",         "rest.{}.@smooth".format(template_name)),
                                                 ("wfmri_output.wavg_epi",       "rest.{}.@avg_epi".format(template_name)),
                                                 ("wfmri_output.warp_field",     "rest.{}.@warp_field".format(template_name)),
                                                ]),
                    ])

    if not do_group_template:
        main_wf.connect([
                         (anat_wf, warp_fmri_wf, [("anat_output.anat_biascorr",  "wfmri_input.reference_file"),
                                                  ("anat_output.warp_forward",   "wfmri_input.anat_to_mni_warp"),
                                                 ]),
                        # output
                        (warp_fmri_wf,  datasink,  [
                                                    ("wfmri_output.coreg_avg_epi", "rest.@coreg_fmri_anat"),
                                                    ("wfmri_output.coreg_others",  "rest.@coreg_others"),
                                                   ]),
                        ])

    if do_atlas:
        main_wf.connect([(anat_wf,      warp_fmri_wf, [("anat_output.atlas_anat",   "wfmri_input.atlas_anat")]),
                         (warp_fmri_wf, datasink,     [("wfmri_output.atlas_fmri",  "rest.@atlas")]),
                         ])
    return main_wf



# -*- coding: utf-8 -*-
"""
PET group template registration nipype workflow.
"""
import os.path as op

import nipype.pipeline.engine as pe
from   nipype.interfaces.io import DataSink
from   nipype.interfaces import IdentityInterface

from   .mrpet import attach_spm_mrpet_preprocessing
from   ..preproc import (spm_create_group_template_wf,
                         spm_register_to_template_wf,)
from   ..config  import setup_node, get_config_setting
from   .._utils  import format_pair_list, flatten_list
from   ..utils   import (get_datasink,
                         extend_trait_list,
                         get_input_node,
                         remove_ext,
                         get_input_file_name,
                         extension_duplicates)


def attach_spm_pet_grouptemplate(main_wf, wf_name="spm_pet_template"):
    """ Attach a PET pre-processing workflow that uses SPM12 to `main_wf`.
    This workflow picks all spm_pet_preproc outputs 'pet_output.warped_files' in `main_wf`
    to create a group template.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    pet_output.warped_files: input node

    datasink: nipype Node

    spm_pet_preproc: nipype Workflow

    Nipype Outputs
    --------------
    group_template.pet_template: file
        The path to the PET group template.

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - spm_pet_preproc
    - spm_anat_preproc if `spm_pet_template.do_petpvc` is True.

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    pet_wf = main_wf.get_node("spm_pet_preproc")

    in_files = get_input_node(main_wf)
    datasink = get_datasink(main_wf, name='datasink')

    # The base name of the 'pet' file for the substitutions
    pet_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'pet')))

    # the group template datasink
    base_outdir  = datasink.inputs.base_directory
    grp_datasink = pe.Node(DataSink(parameterization=False,
                                    base_directory=base_outdir,),
                                    name='{}_grouptemplate_datasink'.format(pet_fbasename))
    grp_datasink.inputs.container = '{}_grouptemplate'.format(pet_fbasename)

    # the list of the raw pet subjects
    warped_pets = pe.JoinNode(interface=IdentityInterface(fields=["warped_pets"]),
                              joinsource="infosrc",
                              joinfield="warped_pets",
                              name="warped_pets")

    # the group template workflow
    template_wf = spm_create_group_template_wf(wf_name)

    # output node
    output = setup_node(IdentityInterface(fields=["pet_template"]), name="group_template")

    # group dataSink output substitutions
    regexp_subst = [
                     (r"/wgrptemplate{pet}_merged_mean_smooth.nii$",  "/{pet}_grouptemplate_mni.nii"),
                     (r"/w{pet}_merged_mean_smooth.nii$",             "/{pet}_grouptemplate_mni.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    grp_datasink.inputs.regexp_substitutions = extend_trait_list(grp_datasink.inputs.regexp_substitutions,
                                                                 regexp_subst)

    # Connect the nodes
    main_wf.connect([
                     # warped pets file list input
                     (pet_wf,       warped_pets, [("warp_output.warped_files",    "warped_pets")]),

                     # group template wf
                     (warped_pets,  template_wf, [(("warped_pets", flatten_list), "grptemplate_input.in_files")]),

                     # output node
                     (template_wf, output,       [("grptemplate_output.template", "pet_template")]),

                     # template output
                     (output,      grp_datasink, [("pet_template",                "@pet_group_template")]),
                   ])

    # Now we start with the correction and registration of each subject to the group template
    do_petpvc = get_config_setting('spm_pet_template.do_petpvc')
    if do_petpvc:
        if main_wf.get_node('spm_anat_preproc') is None:
            raise AttributeError("Expected `spm_anat_preproc` workflow node to attach PETPVC.")

        preproc_wf_name = "spm_mrpet_grouptemplate_preproc"
        main_wf = attach_spm_mrpet_preprocessing(main_wf, wf_name=preproc_wf_name, do_group_template=True)

        preproc_wf = main_wf.get_node(preproc_wf_name)
        main_wf.connect([(output, preproc_wf, [("pet_template", "pet_input.pet_template".format(preproc_wf_name))]), ])
    else:
        # add the pet template to the preproc workflow
        reg_wf = spm_register_to_template_wf(wf_name="spm_pet_register_to_grouptemplate")
        main_wf.connect([
                         (output,   reg_wf,  [("pet_template",  "reg_input.template")]),
                         (in_files, reg_wf,  [("pet",           "reg_input.in_file"),]),

                         (reg_wf,   datasink, [("reg_output.warped",     "pet.grp_template.@warped"),
                                               ("reg_output.warp_field", "pet.grp_template.@warp_field"),
                                              ]),
                         ])

    # per-subject datasink output substitutions
    regexp_subst = [
                     (r"/{pet}_sn.mat$",           "/{pet}_grptemplate_params.mat"),
                     (r"/wgrptemplate_{pet}.nii$", "/{pet}_grptemplate.nii"),
                     (r"/w{pet}.nii",              "/{pet}_grptemplate.nii"),
                   ]
    regexp_subst = format_pair_list(regexp_subst, pet=pet_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    datasink.inputs.regexp_substitutions = extend_trait_list(datasink.inputs.regexp_substitutions,
                                                             regexp_subst)

    return main_wf




# -*- coding: utf-8 -*-
"""
Nipype workflows to use Camino for tractography.
"""
import nipype.pipeline.engine    as pe
import nipype.algorithms.misc    as misc
from   nipype.interfaces.utility import IdentityInterface
from   nipype.interfaces.camino  import (Image2Voxel, FSL2Scheme, DTIFit, Track,
                                         Conmat, ComputeFractionalAnisotropy, AnalyzeHeader)

from   ..config  import setup_node, check_atlas_file
from   ..utils import (get_datasink,
                       get_interface_node,
                       get_input_node,
                       get_data_dims,
                       get_vox_dims,
                       get_affine,
                       )


def camino_tractography(wf_name="camino_tract"):
    """ Run the diffusion MRI pre-processing workflow against the diff files in `data_dir`.

    Nipype Inputs
    -------------
    tract_input.diff: traits.File
        path to the diffusion MRI image

    tract_input.bval: traits.File
        path to the bvals file

    tract_input.bvec: traits.File
        path to the bvecs file

    tract_input.mask: traits.File
        path to the brain mask file

    tract_input.atlas: traits.File
        path to the atlas file

    Nipypte Outputs
    ---------------
    tract_output.tensor
        The result of fitting the tensor model to the whole image.

    tract_output.tracks
        The tractography result.

    tract_output.connectivity
        The atlas ROIxROI structural connectivity matrix.

    tract_output.mean_fa
        The atlas ROIxROI structural connectivity matrix with average FA values.

    tract_output.fa
        The voxelwise fractional anisotropy image.

    Returns
    -------
    wf: nipype Workflow
    """
    in_fields  = ["diff", "bvec", "bval", "mask", "atlas"]
    out_fields = ["tensor", "tracks", "connectivity", "mean_fa", "fa"]

    tract_input  = pe.Node(IdentityInterface(fields=in_fields,
                                             mandatory_inputs=True),
                           name="tract_input")

    img2vox_diff = setup_node(Image2Voxel(out_type="float"), name="img2vox_diff")
    img2vox_mask = setup_node(Image2Voxel(out_type="short"), name="img2vox_mask")
    fsl2scheme   = setup_node(FSL2Scheme(),                  name="fsl2scheme")
    dtifit       = setup_node(DTIFit(),                      name="dtifit")
    fa           = setup_node(ComputeFractionalAnisotropy(), name="fa")

    analyzehdr_fa = setup_node(interface=AnalyzeHeader(), name="analyzeheader_fa")
    analyzehdr_fa.inputs.datatype = "double"

    fa2nii = setup_node(interface=misc.CreateNifti(), name='fa2nii')

    track  = setup_node(Track(inputmodel="dt", out_file="tracts.Bfloat"), name="track")
    conmat = setup_node(Conmat(output_root="conmat_"), name="conmat")

    tract_output = pe.Node(IdentityInterface(fields=out_fields),
                           name="tract_output")

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                # convert data to camino format
                (tract_input,   img2vox_diff,     [("diff",                  "in_file"     )]),
                (tract_input,   img2vox_mask,     [("mask",                  "in_file"     )]),

                # convert bvec and bval to camino scheme
                (tract_input,   fsl2scheme,       [("bvec",                  "bvec_file"   ),
                                                   ("bval",                  "bval_file"   )]),

                # dtifit
                (img2vox_diff,  dtifit,           [("voxel_order",           "in_file"     )]),
                (img2vox_mask,  dtifit,           [("voxel_order",           "bgmask"      )]),
                (fsl2scheme,    dtifit,           [("scheme",                "scheme_file" )]),

                # calculate FA
                (fsl2scheme,    fa,               [("scheme",                "scheme_file" )]),
                (dtifit,        fa,               [("tensor_fitted",         "in_file"     )]),

                # tractography
                (tract_input,   track,            [("atlas",                 "seed_file"   )]),
                (dtifit,        track,            [("tensor_fitted",         "in_file"     )]),

                # convert FA data to NifTI
                (fa,            analyzehdr_fa,    [("fa",                    "in_file"     )]),
                (tract_input,   analyzehdr_fa,    [(('diff', get_vox_dims),  "voxel_dims"  ),
                                                   (('diff', get_data_dims), "data_dims"   )]),

                (tract_input,   fa2nii,           [(("diff", get_affine),    "affine"      )]),
                (analyzehdr_fa, fa2nii,           [("header",                "header_file" )]),
                (fa,            fa2nii,           [("fa",                    "data_file"   )]),

                # connectivity matrix
                (tract_input,   conmat,           [("atlas",                 "target_file" )]),
                (track,         conmat,           [("tracked",               "in_file"     )]),
                (fa2nii,        conmat,           [("nifti_file",            "scalar_file" )]),

                # output
                (fa2nii,        tract_output,     [("nifti_file",            "fa"          )]),
                (dtifit,        tract_output,     [("tensor_fitted",         "tensor"      )]),
                (track,         tract_output,     [("tracked",               "tracks"      )]),
                (conmat,        tract_output,     [("conmat_sc",             "connectivity"),
                                                   ("conmat_ts",             "mean_fa"     )]),
              ])
    return wf


def attach_camino_tractography(main_wf, wf_name="camino_tract"):
    """ Attach the Camino-based tractography workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    atlas_file: str
        Path to the anatomical atlas.

    wf_name: str
        Name of the preprocessing workflow

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.select.diff: input node

    datasink: nipype Node

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - spm_anat_preproc
    - spm_fsl_dti_preprocessing

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)
    dti_coreg_output = get_interface_node(main_wf, 'dti_co_output')
    dti_artif_output = get_interface_node(main_wf, 'dti_art_output')

    # The workflow box
    tract_wf = camino_tractography(wf_name=wf_name)

    # input and output diffusion MRI workflow to main workflow connections
    main_wf.connect([(in_files,         tract_wf, [("bval",            "tract_input.bval")]),
                     (dti_coreg_output, tract_wf, [("brain_mask_diff", "tract_input.mask")]),

                     (dti_artif_output, tract_wf, [("eddy_corr_file",  "tract_input.diff"),
                                                   ("bvec_rotated",    "tract_input.bvec"),
                                                  ]),

                     # output
                     (tract_wf, datasink, [("tract_output.tensor",       "tract.@tensor"),
                                           ("tract_output.tracks",       "tract.@tracks"),
                                           ("tract_output.connectivity", "tract.@connectivity"),
                                           ("tract_output.mean_fa",      "tract.@mean_fa"),
                                           ("tract_output.fa",           "tract.@fa"),
                                           ])
                    ])

    # pass the atlas if it's the case
    do_atlas, _ = check_atlas_file()
    if do_atlas:
        main_wf.connect([(dti_coreg_output, tract_wf, [("atlas_diff", "tract_input.atlas")])])

    return main_wf



# -*- coding: utf-8 -*-
"""
A nipype.SelectFiles node based on hansel.Crumb
"""
import os.path as op
from warnings import warn

from hansel import Crumb

from nipype.interfaces.base import (traits,
                                    DynamicTraitedSpec,
                                    Undefined, BaseInterfaceInputSpec)
from nipype.interfaces.io import IOBase, add_traits
from nipype.utils.filemanip import list_to_filename
from nipype.utils.misc import human_order_sorted

from ._utils import get_values_map_keys


class DataCrumbInputSpec(DynamicTraitedSpec, BaseInterfaceInputSpec):
    sort_filelist = traits.Bool(True, usedefault=True,
                                desc='Sort the filelist that matches the template. Crumb always sort its outputs.')
    raise_on_empty = traits.Bool(True, usedefault=True,
                                 desc="Raise an exception if a template pattern matches no files.")
    force_lists = traits.Either(traits.Bool(), traits.List(traits.Str()),
                                default=False, usedefault=True,
                                desc=("Whether to return outputs as a list even when only one file "
                                      "matches the template. Either a boolean that applies to all "
                                      "output fields or a list of output field names to coerce to "
                                      " a list"))


class DataCrumb(IOBase):
    """ Use Crumb from hansel to select input files."""
    input_spec  = DataCrumbInputSpec
    output_spec = DynamicTraitedSpec
    _always_run = True

    def __init__(self, crumb, templates, **kwargs):
        """Create an instance with specific input fields.
        Parameters
        ----------
        crumb: hansel.Crumb
            If you are using a relative crumb path use a first argument as
            base directory. This argument will be exposed as an input.
            Example:
            {base_dir}/data/raw/{subj_id}...

        templates : dict[str] -> list of 2-tuples
            Mapping from string keys to list of crumb arguments in crumb_path
            that must be replaced to complete the file crumb path.

            The keys become output fields on the interface.

            At runtime, the values of the interface inputs will be
            plugged into these templates, and the resulting strings will be
            used to select files.
        """
        super(DataCrumb, self).__init__(**kwargs)

        # Infer the infields and outfields from the template
        if not Crumb.is_valid(crumb):
            raise ValueError('Crumb {} is not valid.'.format(crumb))

        self._crumb = crumb

        files_args = get_values_map_keys(templates)
        undef_args = [name for name in list(crumb.all_args()) if name not in files_args]
        self._infields  = undef_args

        self._outfields = []
        self._templates = []
        if templates:
            self._outfields = list(templates)
            self._templates = templates

        # Add the dynamic input fields
        undefined_traits = {}
        for field in self._infields:
            self.inputs.add_trait(field, traits.Any)
            undefined_traits[field] = Undefined
        self.inputs.trait_set(trait_change_notify=False, **undefined_traits)

    def _add_output_traits(self, base):
        """Add the dynamic output fields"""
        out_fields = list(self._templates.keys()) + list(self._infields)
        return add_traits(base, out_fields)

    def _list_outputs(self):
        """Find the files and expose them as interface outputs."""
        outputs = {}
        info = dict([(k, v) for k, v in list(self.inputs.__dict__.items())
                     if k in self._infields])

        # check if the crumb is not absolute or if in info we have the parameter for the base directory
        if not self._crumb.isabs():
            first_arg_name, _ = self._crumb._first_open_arg()
            if first_arg_name not in info:
                raise KeyError('Crumb path is not absolute and could not find input for {}.'.format(first_arg_name))
            elif not op.isabs(info[first_arg_name]):
                raise IOError('Expected an absolute path for {} argument in {} but got {}.'.format(first_arg_name,
                                                                                                   self._crumb,
                                                                                                   info[first_arg_name],
                                                                                                   ))
        force_lists = self.inputs.force_lists
        if isinstance(force_lists, bool):
            force_lists = self._outfields if force_lists else []
        bad_fields = set(force_lists) - set(self._outfields)
        if bad_fields:
            bad_fields = ", ".join(list(bad_fields))
            plural = "s" if len(bad_fields) > 1 else ""
            verb = "were" if len(bad_fields) > 1 else "was"
            msg = ("The field%s '%s' %s set in 'force_lists' and not in "
                   "'templates'.") % (plural, bad_fields, verb)
            raise ValueError(msg)

        # loop over the crumb arguments to fill self_crumb
        crumb_info = {k: v for k, v in info.items() if k in self._crumb.open_args()}
        ocrumb = self._crumb.replace(**crumb_info)

        # check again if crumb path is absolute
        if not ocrumb.isabs():
            raise ValueError('Expected a Crumb with an absolute path, got {}.'.format(ocrumb))

        if not ocrumb.exists():
            raise IOError('Expected an existing Crumb path, got {}.'.format(ocrumb))

        # loop over all the ouput items and fill them with the info in templates
        for field, template in self._templates.items():

            # Fill in the template and glob for files
            focrumb  = ocrumb.replace(**dict(template))

            if list(focrumb.open_args()):
                raise ValueError('Expected a full specification of the Crumb path by now, got {}.'.format(focrumb))

            filelist = [cr.path for cr in focrumb.unfold()]
            # Handle the case where nothing matched
            if not filelist:
                msg = "No files were found unfolding %s crumb path: %s" % (
                    field, focrumb)
                if self.inputs.raise_on_empty:
                    raise IOError(msg)
                else:
                    warn(msg)

            # Possibly sort the list
            if self.inputs.sort_filelist:
                filelist = human_order_sorted(filelist)

            # Handle whether this must be a list or not
            if field not in force_lists:
                filelist = list_to_filename(filelist)

            outputs[field] = filelist

            # add the crumb argument values for output
            for arg_name in focrumb.all_args():
                outputs[arg_name] = focrumb[arg_name][0]

        return outputs



# -*- coding: utf-8 -*-
"""
Nipype interfaces to perform signal decomposition.
"""

import nipype.pipeline.engine as pe
from   nipype.interfaces.io import DataSink
from   nipype.interfaces import IdentityInterface, Function

from   .._utils import _check_list
from   ..config import setup_node, get_config_setting
from   ..interfaces import CanICAInterface
from   ..interfaces.nilearn.image import concat_3D_imgs
from   ..utils import (get_trait_value,
                       get_datasink, )
from   .plotting import plot_ica_results


def attach_canica(main_wf, wf_name="canica", **kwargs):
    """ Attach a nilearn CanICA interface to `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    kwargs: dict[str]->str
        input_node: str
            Name of the input node from where to connect the source `input_connect`.

        input_connection: str
            Name of the connection to obtain the source files.

    Nipype Inputs for `main_wf`
    ---------------------------
    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    srcwf_name   = kwargs['input_node']
    srcconn_name = kwargs['input_connection']

    src_wf   = main_wf.get_node(srcwf_name)
    datasink = get_datasink(main_wf, name='datasink')

    base_outdir  = datasink.inputs.base_directory
    ica_datasink = pe.Node(DataSink(parameterization=False,
                                    base_directory=base_outdir,),
                           name="{}_datasink".format(wf_name))

    # the list of the subjects files
    ica_subjs = pe.JoinNode(interface=IdentityInterface(fields=["ica_subjs"]),
                            joinsource="infosrc",
                            joinfield="ica_subjs",
                            name="ica_subjs")

    # warp each subject to the group template
    ica = setup_node(CanICAInterface(), name="{}_ica".format(wf_name),)

    # Connect the nodes
    main_wf.connect([
                     # file list input
                     (src_wf,     ica_subjs, [(srcconn_name, "ica_subjs")]),

                     # canica
                     (ica_subjs,  ica,    [("ica_subjs",  "in_files")]),

                     # canica output
                     (ica, ica_datasink,  [("components", "canica.@components")]),
                     (ica, ica_datasink,  [("loadings",   "canica.@loadings")]),
                     (ica, ica_datasink,  [("score",      "canica.@score")]),
                   ])
    return main_wf


def attach_concat_canica(main_wf, wf_name="canica", **kwargs):
    """ Attach a Concat and a nilearn CanICA interface to `main_wf`.

    The Concat node will merge all the files together in one 4D volume before delivering it to CanICA.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow

    kwargs: dict[str]->str
        input_node: str
            Name of the input node from where to connect the source `input_connect`.

        input_connection: str
            Name of the connection to obtain the source files.

    Nipype Inputs for `main_wf`
    ---------------------------
    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    srcwf_name   = kwargs['input_node']
    srcconn_name = kwargs['input_connection']

    src_wf   = main_wf.get_node(srcwf_name)
    datasink = get_datasink(main_wf, name='datasink')

    base_outdir  = datasink.inputs.base_directory
    ica_datasink = pe.Node(DataSink(parameterization=False,
                                    base_directory=base_outdir,),
                           name="ica_datasink".format(wf_name))
    ica_datasink.inputs.container = 'ica_{}'.format(wf_name)

    # the list of the raw pet subjects
    ica_subjs = pe.JoinNode(interface=IdentityInterface(fields=["ica_subjs"]),
                            joinsource="infosrc",
                            joinfield="ica_subjs",
                            name="ica_subjs")

    # concat images
    concat = setup_node(Function(function=concat_3D_imgs,
                                 input_names=["in_files"],
                                 output_names=["out_file"],
                                 imports=['from neuro_pypes.interfaces.nilearn import ni2file']),
                        name="concat")

    # warp each subject to the group template
    ica = setup_node(CanICAInterface(), name="{}_ica".format(wf_name),)
    algorithm = get_config_setting("{}_ica.algorithm".format(wf_name),
                                   default=get_config_setting('canica.algorithm',
                                   default=''))
    if algorithm:
        ica.inputs.algorithm = algorithm

    # Connect the nodes
    main_wf.connect([
                     # file list input
                     (src_wf, ica_subjs, [(srcconn_name, "ica_subjs")]),

                     # concat images
                     (ica_subjs, concat, [("ica_subjs", "in_files")]),

                     # canica
                     (concat, ica, [(("out_file", _check_list), "in_files")]),

                     # canica output
                     (ica, ica_datasink, [("components", "@components"),
                                          ("loadings",   "@loadings"),
                                          ("score",      "@score"),
                                         ]),
                   ])

    # plot the ICA results?
    do_plot = get_config_setting('canica_extra.plot', default=True)
    if not do_plot:
        return main_wf

    # get the plot threshold from the ICA node or the config file (in that order).
    plot_thr = get_config_setting('canica_extra.plot_thr', default=0)
    plot_thr = get_trait_value(ica.inputs, 'threshold', default=plot_thr)

    # plto ica results images
    plot_ica = setup_node(Function(function=plot_ica_results,
                                   input_names=["ica_result", "application", "mask_file", "zscore", "bg_img"],
                                   output_names=["all_icc_plot", "iccs_plot", "sliced_ic_plots"],),
                          name="plot_ica")
    plot_ica.inputs.zscore      = plot_thr
    plot_ica.inputs.mask_file   = get_trait_value(ica.inputs, 'mask')
    plot_ica.inputs.application = 'nilearn'

    # Connect the plotting nodes
    main_wf.connect([
                     # canica
                     (ica,   plot_ica,        [("components",   "ica_result")]),

                     # canica output
                     (plot_ica, ica_datasink, [("all_icc_plot",     "@all_icc_plot"),
                                               ("iccs_plot",        "@iccs_plot"),
                                               ("sliced_ic_plots",  "@sliced_ic_plots"),
                                              ]),
                     ])

    return main_wf


# -*- coding: utf-8 -*-
"""
DICOM to Nifti converter node based on dcm2niix and hansel.Crumb.
"""
import nipype.pipeline.engine as pe
from nipype.interfaces.base    import traits
from nipype.interfaces.utility import IdentityInterface
from nipype.interfaces.dcm2nii import Dcm2niix, Dcm2nii

from ..config import setup_node
from ..utils  import (get_datasink,
                      get_input_node)


def dcm2niix_wf(wf_name='dcm2niix'):
    """Run dcm2niix over one folder with DICOM files.

    Nipype Inputs
    -------------
    dcm2niix.in_dcmdir: traits.Dir
        path to the DICOM images folder.

    Nipype Outputs
    --------------
    dcm2niix.bids: (a list of items which are an existing file name)

    dcm2niix.bvals: (a list of items which are an existing file name)

    dcm2niix.bvecs: (a list of items which are an existing file name)

    dcm2niix.converted_files: (a list of items which are an existing file name)

    Returns
    -------
    wf: nipype Workflow
    """
    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # specify input and output fields
    in_fields  = ["in_dcmdir",]

    out_fields = ["bids",
                  "bvals",
                  "bvecs",
                  "converted_files"]

    # input node
    dcm2niix_input = setup_node(IdentityInterface(fields=in_fields,
                                                  mandatory_inputs=True),
                                name="dcm2niix_input")
    # T1 preprocessing nodes
    dcm2niix = setup_node(Dcm2niix(), name="dcm2niix")

    # output node
    dcm2niix_output = setup_node(IdentityInterface(fields=out_fields),
                                 name="dcm2niix_output")

    # Connect the nodes
    wf.connect([
                # input
                (dcm2niix_input, dcm2niix, [("in_dcmdir",  "source_dir"),
                                           ]),

                # output
                (dcm2niix, dcm2niix_output, [("bids",            "bids"),
                                             ("bvals",           "bvals"),
                                             ("bvecs",           "bvecs"),
                                             ("converted_files", "converted_files"),
                                            ]),
              ])

    return wf


def attach_dcm2niix(main_wf, wf_name="dcm2niix"):
    """ Attach the dcm2niix workflow to the `main_wf`.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the dcm2niix workflow

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    input_files.dcm_dir: input node

    datasink: nipype Node

    Returns
    -------
    main_wf: nipype Workflow
    """
    in_files = get_input_node(main_wf)
    datasink = get_datasink  (main_wf)

    # The workflow box
    d2n_wf = dcm2niix_wf(wf_name=wf_name)

    main_wf.connect([(in_files, d2n_wf,   [("dcm_dir",                         "dcm2niix_input.in_dcmdir")]),
                     (d2n_wf,   datasink, [("dcm2niix_output.bids",            "@bids"),
                                           ("dcm2niix_output.bvals",           "@bvals"),
                                           ("dcm2niix_output.bvecs",           "@bvecs"),
                                           ("dcm2niix_output.converted_files", "@converted_files"),
                                          ],),
                    ])

    return main_wf


def dcm2nii_converter(source_names=traits.Undefined):
    """ Create a dcm2nii workflow.

    It does:
    - dcm2nii

    Nipype Inputs
    -------------
    source_names: traits.File
        path to the DICOM files or series folder.

    Returns
    -------
    wf: nipype Workflow

    Note
    ----
    For more info: http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.dcm2nii.html
    # TODO: an attach function for this node.
    """
    dcm2nii = Dcm2nii()

    dcm2nii.inputs.gzip_output     = True
    dcm2nii.inputs.output_dir      = '.'
    dcm2nii.inputs.terminal_output = 'file'
    dcm2nii.inputs.source_names    = source_names

    return dcm2nii




# -*- coding: utf-8 -*-
"""
fMRI group template registration nipype workflow.
"""
import os.path as op

import nipype.pipeline.engine as pe
from   nipype.interfaces.io import DataSink
from   nipype.interfaces import IdentityInterface

from   ..preproc import spm_create_group_template_wf, spm_warp_to_mni
from   ..config  import setup_node
from   .._utils  import format_pair_list
from   ..utils   import (get_datasink,
                         extend_trait_list,
                         get_input_node,
                         remove_ext,
                         get_input_file_name,
                         extension_duplicates)


def attach_spm_fmri_grouptemplate_wf(main_wf, wf_name='spm_epi_grouptemplate'):
    """ Attach a fMRI pre-processing workflow that uses SPM12 to `main_wf`.
    This workflow picks all spm_fmri_preproc outputs 'fmri_output.warped_files' in `main_wf`
    to create a group template.

    Parameters
    ----------
    main_wf: nipype Workflow

    wf_name: str
        Name of the preprocessing workflow


    Nipype Inputs
    -------------
    rest_input.in_file: traits.File
        The slice time and motion corrected fMRI file.

    Nipype Inputs for `main_wf`
    ---------------------------
    Note: The `main_wf` workflow is expected to have an `input_files` and a `datasink` nodes.

    rest_output.avg_epi_mni: input node

    datasink: nipype Node

    spm_rest_preproc_mni: nipype Workflow

    Nipype Outputs
    --------------
    group_template.fmri_template: file
        The path to the fMRI group template.

    Nipype Workflow Dependencies
    ----------------------------
    This workflow depends on:
    - fmri_cleanup: for the `rest_output.avg_epi` output

    Returns
    -------
    main_wf: nipype Workflow
    """
    # Dependency workflows
    fmri_cleanup_wf = main_wf.get_node('fmri_cleanup')

    in_files = get_input_node(main_wf)
    datasink = get_datasink(main_wf, name='datasink')

    # The base name of the 'rest' file for the substitutions
    fmri_fbasename = remove_ext(op.basename(get_input_file_name(in_files, 'rest')))

    # the group template datasink
    base_outdir  = datasink.inputs.base_directory
    grp_datasink = pe.Node(DataSink(parameterization=False,
                                    base_directory=base_outdir,),
                                    name='{}_grouptemplate_datasink'.format(fmri_fbasename))
    grp_datasink.inputs.container = '{}_grouptemplate'.format(fmri_fbasename)

    # the list of the average EPIs from all the subjects
    #avg_epi_map = pe.MapNode(IdentityInterface(fields=['avg_epis']), iterfield=['avg_epis'], name='avg_epi_map')

    avg_epis = pe.JoinNode(IdentityInterface(fields=['avg_epis']), joinsource='infosrc', joinfield='avg_epis',
                           name='avg_epis')

    # directly warp the avg EPI to the SPM standard template
    warp_epis = spm_warp_to_mni("spm_warp_avgepi_to_mni")

    # the group template workflow
    template_wf = spm_create_group_template_wf(wf_name)

    # output node
    output = setup_node(IdentityInterface(fields=['fmri_template']), name='group_template')

    # group dataSink output substitutions
    regexp_subst = [
                     (r'/wgrptemplate{fmri}_merged_mean_smooth.nii$',  '/{fmri}_grouptemplate_mni.nii'),
                     (r'/w{fmri}_merged_mean_smooth.nii$',             '/{fmri}_grouptemplate_mni.nii'),
                   ]
    regexp_subst = format_pair_list(regexp_subst, fmri=fmri_fbasename)
    regexp_subst += extension_duplicates(regexp_subst)
    grp_datasink.inputs.regexp_substitutions = extend_trait_list(grp_datasink.inputs.regexp_substitutions,
                                                                 regexp_subst)

    # Connect the nodes
    main_wf.connect([
                     # the avg EPI inputs
                     (fmri_cleanup_wf, avg_epis,  [('rest_output.avg_epi',         'avg_epis')]),
                     #(fmri_cleanup_wf, avg_epi_map,  [('rest_output.avg_epi',         'avg_epis')]),
                     #(avg_epi_map,     avg_epis,     [('avg_epis',                    'avg_epis')]),

                     # warp avg EPIs to MNI
                     (avg_epis,        warp_epis,    [('avg_epis',                    'warp_input.in_files')]),

                     # group template wf
                     (warp_epis,       template_wf,  [('warp_output.warped_files',    'grptemplate_input.in_files')]),

                     # output node
                     (template_wf,     output,       [('grptemplate_output.template', 'fmri_template')]),

                     # template output
                     (output,          grp_datasink, [('fmri_template',               '@fmri_group_template')]),
                     (warp_epis,       grp_datasink, [('warp_output.warped_files',    'individuals.@warped')]),
                   ])

    return main_wf




# -*- coding: utf-8 -*-

from .environ import  spm_tpm_priors_path

from .files   import  (remove_ext,
                       rename,
                       get_extension,
                       get_affine,
                       get_data_dims,
                       get_vox_dims,
                       fetch_one_file,
                       extension_duplicates,)

from .piping  import  (extend_trait_list,
                       selectindex,
                       get_trait_value,
                       fsl_merge,
                       get_node,
                       joinstrings,
                       find_wf_node,
                       get_datasink,
                       get_input_node,
                       get_interface_node,
                       get_input_file_name,)

from .pandas import (add_table_headers,
                     write_tabbed_excel,)




# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.  This
was written to work with FSL version 5.0.4.

    Change directory to provide relative paths for doctests
    >>> import os
    >>> filepath = os.path.dirname(os.path.realpath(__file__))
    >>> datadir = os.path.realpath(os.path.join(filepath,
    ...                            '../../testing/data'))
    >>> os.chdir(datadir)
"""
from __future__ import print_function, division, unicode_literals, absolute_import
from builtins import str

import os
import warnings

from nipype.interfaces.base import (traits, TraitedSpec, File, isdefined)
from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec


class EddyInputSpec(FSLCommandInputSpec):
    in_file = File(exists=True, mandatory=True, argstr='--imain=%s',
                   desc=('File containing all the images to estimate '
                         'distortions for'))
    in_mask = File(exists=True, mandatory=True, argstr='--mask=%s',
                   desc='Mask to indicate brain')
    in_index = File(exists=True, mandatory=True, argstr='--index=%s',
                    desc=('File containing indices for all volumes in --imain '
                          'into --acqp and --topup'))
    in_acqp = File(exists=True, mandatory=True, argstr='--acqp=%s',
                   desc='File containing acquisition parameters')
    in_bvec = File(exists=True, mandatory=True, argstr='--bvecs=%s',
                   desc=('File containing the b-vectors for all volumes in '
                         '--imain'))
    in_bval = File(exists=True, mandatory=True, argstr='--bvals=%s',
                   desc=('File containing the b-values for all volumes in '
                         '--imain'))
    out_base = traits.Str('eddy_corrected', argstr='--out=%s',
                          usedefault=True,
                          desc=('basename for output (warped) image'))
    session = File(exists=True, argstr='--session=%s',
                   desc=('File containing session indices for all volumes in '
                         '--imain'))
    in_topup_fieldcoef = File(exists=True, argstr="--topup=%s",
                              requires=['in_topup_movpar'],
                              desc=('topup file containing the field '
                                    'coefficients'))
    in_topup_movpar = File(exists=True, requires=['in_topup_fieldcoef'],
                           desc='topup movpar.txt file')

    flm = traits.Enum('linear', 'quadratic', 'cubic', argstr='--flm=%s',
                      desc='First level EC model')

    fwhm = traits.Float(desc=('FWHM for conditioning filter when estimating '
                              'the parameters'), argstr='--fwhm=%s')

    niter = traits.Int(5, argstr='--niter=%s', desc='Number of iterations')

    method = traits.Enum('jac', 'lsr', argstr='--resamp=%s',
                         desc=('Final resampling method (jacobian/least '
                               'squares)'))
    repol = traits.Bool(False, argstr='--repol',
                        desc='Detect and replace outlier slices')
    num_threads = traits.Int(1, usedefault=True, nohash=True,
                             desc="Number of openmp threads to use")
    is_shelled = traits.Bool(False, argstr='--data_is_shelled',
                             desc="Override internal check to ensure that "
                                  "date are acquired on a set of b-value "
                                  "shells")
    field = traits.Str(argstr='--field=%s',
                       desc="NonTOPUP fieldmap scaled in Hz - filename has "
                            "to be provided without an extension. TOPUP is "
                            "strongly recommended")
    field_mat = File(exists=True, argstr='--field_mat=%s',
                     desc="Matrix that specifies the relative locations of "
                          "the field specified by --field and first volume "
                          "in file --imain")
    use_cuda = traits.Bool(False, desc="Run eddy using cuda gpu")


class EddyOutputSpec(TraitedSpec):
    out_corrected = File(exists=True,
                         desc=('4D image file containing all the corrected '
                               'volumes'))
    out_parameter = File(exists=True,
                         desc=('text file with parameters definining the '
                               'field and movement for each scan'))
    out_rotated_bvecs = File(exists=True,
                             desc=('File containing rotated b-values for all volumes'))
    out_movement_rms = File(exists=True,
                            desc=('Summary of the "total movement" in each volume'))
    out_restricted_movement_rms = File(exists=True,
                                       desc=('Summary of the "total movement" in each volume '
                                             'disregarding translation in the PE direction'))
    out_shell_alignment_parameters = File(exists=True,
                                          desc=('File containing rigid body movement parameters '
                                                'between the different shells as estimated by a '
                                                'post-hoc mutual information based registration'))
    out_outlier_report = File(exists=True,
                              desc=('Text-file with a plain language report '
                                    'on what outlier slices eddy has found'))


class Eddy(FSLCommand):
    """
    Interface for FSL eddy, a tool for estimating and correcting eddy
    currents induced distortions. `User guide
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Eddy/UsersGuide>`_ and
    `more info regarding acqp file
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/Faq#How_do_I_know_what_to_put_into_my_--acqp_file>`_.

    Examples
    --------

    >>> from nipype.interfaces.fsl import Eddy
    >>> eddy = Eddy()
    >>> eddy.inputs.in_file = 'epi.nii'
    >>> eddy.inputs.in_mask  = 'epi_mask.nii'
    >>> eddy.inputs.in_index = 'epi_index.txt'
    >>> eddy.inputs.in_acqp  = 'epi_acqp.txt'
    >>> eddy.inputs.in_bvec  = 'bvecs.scheme'
    >>> eddy.inputs.in_bval  = 'bvals.scheme'
    >>> eddy.cmdline # doctest: +ELLIPSIS +ALLOW_UNICODE
    'eddy_openmp --acqp=epi_acqp.txt --bvals=bvals.scheme --bvecs=bvecs.scheme \
--imain=epi.nii --index=epi_index.txt --mask=epi_mask.nii \
--out=.../eddy_corrected'
    >>> res = eddy.run() # doctest: +SKIP

    """
    _cmd = 'eddy'
    input_spec = EddyInputSpec
    output_spec = EddyOutputSpec

    _num_threads = 1

    def __init__(self, **inputs):
        super(Eddy, self).__init__(**inputs)
        self.inputs.on_trait_change(self._num_threads_update, 'num_threads')
        if isdefined(self.inputs.use_cuda):
            self._use_cuda()
        if not isdefined(self.inputs.num_threads):
            self.inputs.num_threads = self._num_threads
        else:
            self._num_threads_update()

    def _num_threads_update(self):
        self._num_threads = self.inputs.num_threads
        if not isdefined(self.inputs.num_threads):
            if 'OMP_NUM_THREADS' in self.inputs.environ:
                del self.inputs.environ['OMP_NUM_THREADS']
        else:
            self.inputs.environ['OMP_NUM_THREADS'] = str(
                self.inputs.num_threads)

    def _use_cuda(self):
        if self.inputs.use_cuda:
            _cmd = 'eddy_cuda'
        else:
            _cmd = 'eddy_openmp'

    def _format_arg(self, name, spec, value):
        if name == 'in_topup_fieldcoef':
            return spec.argstr % value.split('_fieldcoef')[0]
        if name == 'out_base':
            return spec.argstr % os.path.abspath(value)
        return super(Eddy, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_corrected'] = os.path.abspath(
            '%s.nii.gz' % self.inputs.out_base)
        outputs['out_parameter'] = os.path.abspath(
            '%s.eddy_parameters' % self.inputs.out_base)

        # File generation might depend on the version of EDDY
        out_rotated_bvecs = os.path.abspath(
            '%s.eddy_rotated_bvecs' % self.inputs.out_base)
        out_movement_rms = os.path.abspath(
            '%s.eddy_movement_rms' % self.inputs.out_base)
        out_restricted_movement_rms = os.path.abspath(
            '%s.eddy_restricted_movement_rms' % self.inputs.out_base)
        out_shell_alignment_parameters = os.path.abspath(
            '%s.eddy_post_eddy_shell_alignment_parameters' % self.inputs.out_base)
        out_outlier_report = os.path.abspath(
            '%s.eddy_outlier_report' % self.inputs.out_base)

        if os.path.exists(out_rotated_bvecs):
            outputs['out_rotated_bvecs'] = out_rotated_bvecs
        if os.path.exists(out_movement_rms):
            outputs['out_movement_rms'] = out_movement_rms
        if os.path.exists(out_restricted_movement_rms):
            outputs['out_restricted_movement_rms'] = out_restricted_movement_rms
        if os.path.exists(out_shell_alignment_parameters):
            outputs['out_shell_alignment_parameters'] = out_shell_alignment_parameters
        if os.path.exists(out_outlier_report):
            outputs['out_outlier_report'] = out_outlier_report

        return outputs


class EddyCorrectInputSpec(FSLCommandInputSpec):
    in_file = File(exists=True, desc='4D input file', argstr='%s', position=0,
                   mandatory=True)
    out_file = File(desc='4D output file', argstr='%s', position=1,
                    name_source=['in_file'], name_template='%s_edc',
                    output_name='eddy_corrected')
    ref_num = traits.Int(0, argstr='%d', position=2, desc='reference number',
                         mandatory=True, usedefault=True)


class EddyCorrectOutputSpec(TraitedSpec):
    eddy_corrected = File(exists=True,
                          desc='path/name of 4D eddy corrected output file')


class EddyCorrect(FSLCommand):
    """

    .. warning:: Deprecated in FSL. Please use
      :class:`nipype.interfaces.fsl.epi.Eddy` instead

    Example
    -------

    >>> from nipype.interfaces.fsl import EddyCorrect
    >>> eddyc = EddyCorrect(in_file='diffusion.nii',
    ...                     out_file="diffusion_edc.nii", ref_num=0)
    >>> eddyc.cmdline # doctest: +ALLOW_UNICODE
    'eddy_correct diffusion.nii diffusion_edc.nii 0'

    """
    _cmd = 'eddy_correct'
    input_spec = EddyCorrectInputSpec
    output_spec = EddyCorrectOutputSpec

    def __init__(self, **inputs):
        warnings.warn(("Deprecated: Please use nipype.interfaces.fsl.epi.Eddy "
                       "instead"), DeprecationWarning)
        return super(EddyCorrect, self).__init__(**inputs)

    def _run_interface(self, runtime):
        runtime = super(EddyCorrect, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime



# -*- coding: utf-8 -*-
"""
Helper functions to manage external files.
"""
import re
from   os          import path as op
from   glob        import glob
from   functools   import wraps


def get_vox_dims(volume):
    import nibabel as nb
    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume)
    hdr = nii.header
    voxdims = hdr.get_zooms()
    return [float(voxdims[0]), float(voxdims[1]), float(voxdims[2])]


def get_data_dims(volume):
    import nibabel as nb
    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume)
    hdr = nii.header
    datadims = hdr.get_data_shape()
    return [int(datadims[0]), int(datadims[1]), int(datadims[2])]


def get_affine(volume):
    import nibabel as nb
    nii = nb.load(volume)
    return nii.affine


def niftiimg_out(f):
    """ Picks a function whose first argument is an `img` or a sequence of imgs, processes its
    data and returns a numpy array. This decorator wraps this numpy array
    into a nibabel.Nifti1Image."""
    import nibabel as nib
    import nilearn.image as niimg

    @wraps(f)
    def wrapped(*args, **kwargs):
        r = f(*args, **kwargs)

        img = niimg.load_img(args[0])
        return nib.Nifti1Image(r, affine=img.get_affine(), header=img.header)

    return wrapped


def get_extension(filepath, check_if_exists=False, allowed_exts=None):
    """Return the extension of filepath.

    Parameters
    ----------
    filepath: string
        File name or path

    check_if_exists: bool

    allowed_exts: dict
        Dictionary of strings, where the key if the last part of a complex ('.' separated) extension
        and the value is the previous part.
        For example: for the '.nii.gz' extension I would have a dict as {'.gz': '.nii'}
        Default: {'.gz': '.nii'}

    Returns
    -------
    ext: str
        The extension of the file name or path
    """
    if allowed_exts is None:
        allowed_exts = {'.gz': '.nii'}

    try:
        rest, ext = op.splitext(filepath)
        if ext in allowed_exts:
            alloweds = allowed_exts[ext]
            _, ext2 = op.splitext(rest)
            if ext2 in alloweds:
                ext = ext2 + ext
    except:
        raise
    else:
        return ext


def add_extension_if_needed(filepath, ext):
    """Add the extension ext to fpath if it doesn't have it.

    Parameters
    ----------
    filepath: str
        File name or path

    ext: str
        File extension

    Returns
    -------
    filepath: str
        File name or path with extension added, if needed.
    """
    if not filepath.endswith(ext):
        filepath += ext

    return filepath


def remove_ext(filepath):
    """Removes the extension of the file.

    Parameters
    ----------
    filepath: str
        File path or name

    Returns
    -------
    filepath: str
        File path or name without extension
    """
    return filepath[:filepath.rindex(get_extension(filepath))]


def extension_duplicates(regexp_pair_list):
    """ Return a new list with the pair items of `regexp_pair_list`
    have all the '.nii' replaced by '.nii.gz'.
    This is useful for the Datasink regexp_substitutions when
    you don't know/care what extension the output image will have,
    then you put all of them like:
    "(r"/rc1[\w]+_corrected\.nii$",  "/coreg_gm.nii")"

    and then call this function to add the duplicates with modified
    extension of these same pairs.

    Parameters
    ----------
    regexp_pair_list: list of 2-tuple of str

    Returns
    -------
    mod_regexp_pair_list: list of 2-tuple of str
    """
    replace_ext = lambda x: x.replace('.nii', '.nii.gz')
    dups = [(replace_ext(pair[0]), replace_ext(pair[1]))
            for pair in regexp_pair_list
            if '.nii$' in pair[0]]
    return dups


def rename(in_files, suffix=None):
    """Rename all the files in `in_files` adding the `suffix` keeping its
    extension and basedir."""
    import os.path as path
    from nipype.utils.filemanip import (filename_to_list, split_filename,
                                        list_to_filename)
    out_files = []
    for idx, filename in enumerate(filename_to_list(in_files)):
        base, name, ext = split_filename(filename)
        if suffix is None:
            new_name = name + ('_%03d' % idx) + ext
        else:
            new_name = name + suffix + ext

        out_files.append(path.join(base, new_name))

    return list_to_filename(out_files)


def find_files_in(dirpath, file_pattern, pat_type='fnmatch'):
    """ Find files in `dirpath` without recursivity.

    Parameters
    ----------
    dirpath: str
        Folder where to search for file names.

    file_pattern: str
        File pattern to be matched.

    pat_type: str
        The type of pattern in `file_pattern`.
        Choices: 'fnmatch', 're.search', 're.match'.

    Returns
    -------
    files: List[str]
        List of paths to the files that match file_pattern.
        `dirpath` is included in each of its items.
    """
    if pat_type == 'fnmatch':
        files = glob(op.join(dirpath, file_pattern))
    elif pat_type == 're.search':
        regex = re.compile(file_pattern)
        files = [f for f in glob(op.join(dirpath, '*')) if regex.search(f)]
    elif pat_type == 're.match':
        regex = re.compile(file_pattern)
        files = [f for f in glob(op.join(dirpath, '*')) if regex.match(f)]
    else:
        raise ValueError("Expected one of ('fnmatch', 're.search' or 're.match') for "
                         "`pat_type` parameter, got {}.".format(pat_type))

    return files


def fetch_one_file(dirpath, file_pattern, file_extension=None, extra_prefix=None,
                   extra_suffix=None, pat_type='fnmatch'):
    """ Return the unique file path in dirpath that matches fnmatch file_pattern.
    Add the extra_prefix to try the search again, if the file_pattern finds more than one file.

    Parameters
    ----------
    dirpath:

    file_pattern:
        File pattern to be matched.

    file_extension:
        Extension of the file.

    extra_prefix:

    extra_suffix:

    pat_type: str
        The type of pattern in `file_pattern`.
        Choices: 'fnmatch', 're.search', 're.match'.

    Returns
    -------
    file_path: str
        The path to the unique file that matches the conditions inside `dirpath`.
        `dirpath` is included.

    Raises
    ------
    IOError
        If `dirpath` doesn't exist or the list of returned files is empty.

    ValueError
        IF the choice for `pat_type` is not valid.

    """
    if file_extension is not None:
        file_fnmatch = file_pattern + file_extension
    else:
        file_fnmatch = file_pattern

    files = find_files_in(dirpath, file_pattern, pat_type=pat_type)

    if not files:
        raise IOError('Expected at least one file that matched the '
                      'pattern {} in {}.'.format(file_fnmatch, dirpath))

    if len(files) > 1:
        if extra_prefix is None:
            extra_prefix = ''

        if extra_suffix is None:
            extra_suffix = ''

        if not extra_prefix and not extra_suffix:
            raise IOError('Found more than one file that matched the '
                          'pattern {} in {}: {}'.format(file_fnmatch, dirpath, files))

        else:
            # TODO: be careful, this might only work with fnmatch
            return fetch_one_file(dirpath, extra_prefix + file_pattern + extra_suffix,
                                  file_extension=file_extension,
                                  pat_type=pat_type)

    return files[0]


def save_object(obj, filename):
    """ Save `obj` in `filename` as a pickle."""
    import pickle

    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)



#
#
# def dual_regression(y, X):
#     """ Use dual regression approach to compute the spatial maps
#     and time courses.
#
#     Parameters
#     ----------
#     y: np.ndarray
#         Observations in columns (Voxels by timepoints)
#
#     X: np.ndarray
#         design matrix (Voxels by components)
#
#     Returns
#     -------
#     tc: np.ndarray
#         Time courses (Timepoints by components)
#
#     spatial_maps: np.ndarray
#         Spatial maps (Components by voxels)
#     """
#     X =
#
# def spatial_maps_pairwise_regression(imgs1, imgs2, mask_file, method_num=0):
#     """ TODO!!
#
#     Look in icat_dual_regress:
#
#     function [tc, spatial_maps] = icatb_dual_regress(y, X)
#     %% Use dual regression approach to compute the spatial maps and time
#     % courses
#     %
#     % Inputs:
#     % 1. y - Observations in columns (Voxels by time points)
#     % 2. X - design matrix (Voxels by components)
#     %
#     % Outputs:
#     % 1. tc - Time courses (Timepoints by components)
#     % 2. spatial_maps - Spatial maps (Components by voxels)
#     %
#
#     %% First step. Fit model matrix to the data to get time courses.
#     X = icatb_remove_mean(X);
#
#     tc = pinv(X)*icatb_remove_mean(y);
#     tc = tc';
#     clear X;
#
#     % Store mean of timecourses
#     mean_tc = mean(tc);
#
#     % Remove mean of timecourse
#     tc = icatb_remove_mean(tc);
#
#     %% Second step. Fit Time courses at each voxel to get the spatial maps.
#     try
#
#         spatial_maps = pinv(tc)*icatb_remove_mean(y');
#
#     catch
#         %% Use less memory usage way to do regression
#
#         % Initialise spatial maps
#         spatial_maps = zeros(size(tc, 2), size(y, 1));
#
#
#         % Loop over voxels
#         for nVoxel = 1:size(y, 1)
#
#             bold_signal = detrend(y(nVoxel, :), 0);
#
#             spatial_maps(:, nVoxel) = pinv(tc)*bold_signal(:);
#
#             clear bold_signal;
#
#         end
#         % End of loop over voxels
#
#     end
#
#     clear y;
#
#     %% Add mean back to the timecourses
#     tc = tc + (repmat(mean_tc, size(tc, 1), 1));
#
#
#
#     Multiple Regression values of each RSN to each IC map in `ic_imgs` masked by `mask_file`.
#
#     Parameters
#     ----------
#     imgs1: list of niimg-like or 4D niimg-like
#
#     imgs2: list of niimg-like or 4D niimg-like
#
#     mask_file: niimg-like
#
#     method_name: str
#         Valid values for method_name are:
#         From scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan'].
#         From scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev', 'correlation', 'dice', 'hamming',
#                                       'jaccard', 'kulsinski', 'mahalanobis', 'matching', 'minkowski',
#                                       'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener', 'sokalsneath',
#                                       'sqeuclidean', 'yule']
#                                       See the documentation for scipy.spatial.distance for details on these metrics.
#
#     Returns
#     -------
#     corrs: np.ndarray
#         A matrix of shape MxN, where M is len(rsn_imgs) and N is len(ic_imgs).
#         It contains the correlation values.
#     """
#     rsn_img = niimg.load_img(imgs1)
#     ic_img  = niimg.load_img(imgs2)
#
#     n_rsns = rsn_img.shape[-1]
#     n_ics  =  ic_img.shape[-1]
#     corrs = np.zeros((n_rsns, n_ics), dtype=float)
#
#     mask_trnsf = niimg.resample_to_img(mask_file, niimg.index_img(ic_img, 0),
#                                        interpolation='nearest',
#                                        copy=True)
#
#     for rsn_idx, rsn in enumerate(niimg.iter_img(rsn_img)):
#         rsn_transf = niimg.resample_to_img(rsn, niimg.index_img(ic_img, 0), copy=True)
#         rsn_masked = nimask.apply_mask(rsn_transf, mask_trnsf)
#
#         for ic_idx, ic in enumerate(niimg.iter_img(ic_img)):
#             ic_masked = nimask.apply_mask(ic, mask_trnsf)
#             # dist = pairwise_distances(rsn_masked.reshape(1, -1),
#             #                            ic_masked.reshape(1, -1),
#             #                           metric=distance)
#
#             # -----------------------------------------------------------------------------------------------------
#             # CODE TO MODIFY
#
#             # http://stackoverflow.com/questions/17679140/multiple-linear-regression-with-python
#             # http://scikit-learn.org/stable/modules/linear_model.html#ordinary-least-squares
#             import pandas as pd
#             import numpy as np
#
#             path = 'DB2.csv'
#             data = pd.read_csv(path, header=None, delimiter=";")
#
#             data.insert(0, 'Ones', 1)
#             cols = data.shape[1]
#
#             X = data.iloc[:,0:cols-1]
#             y = data.iloc[:,cols-1:cols]
#
#             IdentitySize = X.shape[1]
#             IdentityMatrix= np.zeros((IdentitySize, IdentitySize))
#             np.fill_diagonal(IdentityMatrix, 1)
#
#             #For least squares method you use Numpy's numpy.linalg.lstsq. Here is Pyhton code
#             lamb = 1
#             th = np.linalg.lstsq(X.T.dot(X) + lamb * IdentityMatrix, X.T.dot(y))[0]
#
#             # Also you can use np.linalg.solve tool of numpy:
#             lamb = 1
#             XtX_lamb = X.T.dot(X) + lamb * IdentityMatrix
#             XtY = X.T.dot(y)
#             x = np.linalg.solve(XtX_lamb, XtY);
#
#             # For normal equation method use:
#             lamb = 1
#             xTx = X.T.dot(X) + lamb * IdentityMatrix
#             XtX = np.linalg.inv(xTx)
#             XtX_xT = XtX.dot(X.T)
#             theta = XtX_xT.dot(y)
#
#             # OLS
#             # >>> from sklearn import linear_model
#             # >>> reg = linear_model.LinearRegression()
#             # >>> reg.fit ([[0, 0], [1, 1], [2, 2]], [0, 1, 2])
#             # LinearRegression(copy_X=True, fit_intercept=True, n_jobs=1, normalize=False)
#             # >>> reg.coef_
#             # array([ 0.5,  0.5])
#
#             # Ridge Regression
#             # >>> from sklearn import linear_model
#             # >>> reg = linear_model.Ridge (alpha = .5)
#             # >>> reg.fit ([[0, 0], [0, 0], [1, 1]], [0, .1, 1])
#             # Ridge(alpha=0.5, copy_X=True, fit_intercept=True, max_iter=None,
#             #       normalize=False, random_state=None, solver='auto', tol=0.001)
#             # >>> reg.coef_
#             # array([ 0.34545455,  0.34545455])
#             # >>> reg.intercept_
#             # 0.13636...
#
#             # CODE TO MODIFY
#             # -----------------------------------------------------------------------------------------------------
#
#             # since this is a scalar value
#             dist = dist[0][0]
#
#             # since this is a distance based on correlation, not a correlation value
#             corr = 1 - dist
#
#             # store it
#             corrs[rsn_idx, ic_idx] = corr
#
#     return corrs




# -*- coding: utf-8 -*-
"""
Nipype interfaces to calculate connectivity measures using nilearn.
"""
import os.path as op

import numpy as np
import nilearn.connectome
from   nilearn.input_data     import NiftiMapsMasker, NiftiLabelsMasker
#from   boyle.nifti.roi        import get_roilist_from_atlas
from   nipype.interfaces.base import (BaseInterface,
                                      TraitedSpec,
                                      InputMultiPath,
                                      BaseInterfaceInputSpec,
                                      traits,)

from ...utils import get_trait_value


class ConnectivityCorrelationInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiPath(traits.File(desc="NifTI image file(s) from where to extract the data. \n"
                                               "If more than one (3D volumes), all should be spatially normalized.",
                                          exists=True, mandatory=True))
    atlas_file = traits.File(desc="Atlas image file defining the connectivity ROIs.\n"
                                  "Must be spatially normalized to in_files.",
                             exists=True, mandatory=True)
    atlas_type = traits.Enum("probabilistic", "labels",
                             desc="The type of atlas.",
                             default="labels")

    # masker options
    smoothing_fwhm = traits.Float(desc="If smoothing_fwhm is defined, it gives the full-width half maximum in "
                                       "millimeters of the spatial smoothing to apply to the signal.",)
    standardize = traits.Bool(desc="If standardize is True, the time-series are centered and normed: "
                                   "their mean is put to 0 and their variance to 1 in the time dimension.",
                              default_value=False)
    resampling_target = traits.Enum("mask", "maps", "data", "labels", "",
                                    desc="Gives which image gives the final shape/size. "
                                         "This depends on the `atlas_type`. "
                                         "For 'probabilistic' you must use 'mask', 'maps' or None; while for"
                                         "'labels' you must use 'data', 'labels' or None."
                                         "Have a look on nilearn docs for more information.")

    # connectome options
    kind = traits.Enum ("correlation", "partial correlation", "tangent", "covariance", "precision",
                        desc="The connectivity matrix kind.", default='covariance')


class ConnectivityCorrelationOutputSpec(TraitedSpec):
    connectivity = traits.File(desc="Numpy text file with the connectivity matrix.")
    timeseries   = traits.File(desc="Numpy text file with the time-series or 4th dimension data matrix extracted "
                                    "from the atlas ROIs.")


class ConnectivityCorrelationInterface(BaseInterface):
    """ Nipype Interface to NiLearn methods to calculate connectivity matrices using 4D data and a
    spatially normalized ROI atlas.

    For more information look at: nilearn.connectome.ConnectivityMeasure
    """
    input_spec = ConnectivityCorrelationInputSpec
    output_spec = ConnectivityCorrelationOutputSpec

    def _run_interface(self, runtime):
        atlas_type        = get_trait_value(self.inputs, 'atlas_type')
        conn_kind         = get_trait_value(self.inputs, 'kind',)
        #rois_list         = get_trait_value(self.inputs, 'rois_list',         default=None)
        smoothing_fwhm    = get_trait_value(self.inputs, 'smoothing_fwhm',    default=None)
        standardize       = get_trait_value(self.inputs, 'standardize',       default=None)
        resampling_target = get_trait_value(self.inputs, 'resampling_target', default=None)

        self._time_series_file = op.abspath('conn_timeseries.txt')
        self._conn_mat_file    = op.abspath('connectivity.txt')

        ## TODO: add parameter to choose the ROI labels to be used.
        # if rois_list is None:
        #     rois_vals = get_roilist_from_atlas(self.inputs.atlas_file)
        # else:
        #     try:
        #         rois_vals = np.loadtxt(rois_list, dtype=int)
        #     except:
        #         raise IOError('Error reading ROIs list file {}.'.format(rois_list))

        in_files = self.inputs.in_files
        if len(in_files) == 1:
            in_files = in_files[0]

        if atlas_type == 'probabilistic':
            AtlasMasker = NiftiMapsMasker
        elif atlas_type == 'labels':
            AtlasMasker = NiftiLabelsMasker

        masker = AtlasMasker(self.inputs.atlas_file,
                             standardize=standardize,
                             smoothing_fwhm=smoothing_fwhm,
                             resampling_target=resampling_target,
                             memory='nilearn_cache',
                             verbose=5)

        self._time_series = masker.fit_transform(in_files)

        conn_measure   = nilearn.connectome.ConnectivityMeasure(kind=conn_kind)
        self._conn_mat = conn_measure.fit_transform([self._time_series])

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()

        np.savetxt(self._time_series_file, self._time_series,        fmt='%.10f')
        np.savetxt(self._conn_mat_file,    self._conn_mat.squeeze(), fmt='%.10f')

        outputs['timeseries'  ] = self._time_series_file
        outputs['connectivity'] = self._conn_mat_file
        return outputs




# -*- coding: utf-8 -*-
"""
Workflows to grab input file structures.
"""
import os.path as op
import logging as log

import nipype.pipeline.engine as pe
from   nipype.interfaces.io import DataSink
from   nipype.interfaces.utility import IdentityInterface
from   hansel.utils import joint_value_map, valuesmap_to_dict

from .crumb  import DataCrumb
from .utils  import extend_trait_list, joinstrings
from .       import configuration


def build_crumb_workflow(wfname_attacher, data_crumb, in_out_kwargs, output_dir,
                         cache_dir='', wf_name="main_workflow"):
    """ Returns a workflow for the give `data_crumb` with the attached workflows
    given by `attach_functions`.

    Parameters
    ----------
    wfname_attacher: dict[Str] -> function
        Dictionary with name of the workflow and its corresponding
         attach function that will be in charge of attaching workflows
         to the main input/output workflow.

    data_crumb: hansel.Crumb
        The crumb until the subject files.
        Example: Crumb('/home/hansel/cobre/raw/{subject_id}/session_1/{modality}/{image_file})
        At least one crumb arguments of `data_crumb` must be open,
        this argument will be replaced by the corresponding image name.

    in_out_kwargs: dict with keyword arguments
        This arguments are for the in_out_crumb_wf.
        Mainly 'files_crumb_args' which will declare the values each file
        type the crumb arguments in `data_crumb` must be replaced with.
        Example:
              {'anat':  [('modality', 'anat_1'),
                         ('image',    'mprage.nii.gz')],
              'rest':   [('modality', 'rest_1'),
                         ('image',    'rest.nii.gz')],
              }

    cache_dir: str
        The working directory of the workflow.

    output_dir: str
        The output folder path.

    wf_name: str
        Name of the main workflow.
    """
    if not data_crumb.exists():
        raise IOError("Expected an existing folder for `data_crumb`, got {}.".format(data_crumb))

    if not data_crumb.isabs():
        raise IOError("Expected an absolute Crumb path for `data_crumb`, got {}.".format(data_crumb))

    if not wfname_attacher or wfname_attacher is None:
        raise ValueError("Expected `wfname_attacher` to have at least one function, "
                         "got {}.".format(wfname_attacher))

    #if not in_out_kwargs or in_out_kwargs is None:
    #    raise ValueError("Expected `in_out_kwargs` to have at least the name for sets of parameters for "
    #                     " `data_crumb`, got {}.".format(in_out_kwargs))

    # check some args
    if not cache_dir:
        cache_dir = op.join(op.dirname(output_dir), "wd")

    # print the configuration parameters
    log.info('Using the following configuration parameters:')
    log.info(configuration)

    # generate the workflow
    main_wf = crumb_wf(work_dir=cache_dir,
                       data_crumb=data_crumb,
                       output_dir=output_dir,
                       file_templates=in_out_kwargs,
                       wf_name=wf_name)

    for wf_name, attach_wf in wfname_attacher.items():
        main_wf = attach_wf(main_wf=main_wf, wf_name=wf_name)

    # move the crash files folder elsewhere
    main_wf.config["execution"]["crashdump_dir"] = op.join(main_wf.base_dir,
                                                           main_wf.name, "log")

    log.info('Workflow created.')

    return main_wf


def crumb_wf(work_dir, data_crumb, output_dir, file_templates,
             wf_name="main_workflow"):
    """ Creates a workflow with the `subject_session_file` input nodes and an empty `datasink`.
    The 'datasink' must be connected afterwards in order to work.

    Parameters
    ----------
    work_dir: str
        Path to the workflow temporary folder

    data_crumb: hansel.Crumb
        The crumb until the subject files.
        Example: Crumb('/home/hansel/data/{subject_id}/{session_id}/{modality}/{image_file})

    output_dir: str
        Path to where the datasink will leave the results.

    file_templates: Dict[str -> list of 2-tuple]
        Maps of crumb argument values to specify each file in the `data_crumb`.
        Example: {'anat': [('modality', 'anat'), ('image_file', 'anat_hc.nii.gz')],
                  'pet':  [('modality', 'pet'),  ('image_file', 'pet_fdg.nii.gz')],
                 }

    wf_name: str
        Name of the main workflow

    Returns
    -------
    wf: Workflow
    """
    # create the root workflow
    wf = pe.Workflow(name=wf_name, base_dir=work_dir)

    # datasink
    datasink = pe.Node(DataSink(parameterization=False,
                                base_directory=output_dir,),
                       name="datasink")

    # input workflow
    # (work_dir, data_crumb, crumb_arg_values, files_crumb_args, wf_name="input_files"):
    select_files = pe.Node(DataCrumb(crumb=data_crumb,
                                     templates=file_templates,
                                     raise_on_empty=False),
                           name='selectfiles')

    # basic file name substitutions for the datasink
    undef_args = select_files.interface._infields
    substitutions = [(name, "") for name in undef_args]
    substitutions.append(("__", "_"))

    datasink.inputs.substitutions = extend_trait_list(datasink.inputs.substitutions,
                                                      substitutions)

    # Infosource - the information source that iterates over crumb values map from the filesystem
    infosource = pe.Node(interface=IdentityInterface(fields=undef_args), name="infosrc")
    infosource.iterables = list(valuesmap_to_dict(joint_value_map(data_crumb, undef_args)).items())
    infosource.synchronize = True

    # connect the input_wf to the datasink
    joinpath = pe.Node(joinstrings(len(undef_args)), name='joinpath')

    # Connect the infosrc node to the datasink
    input_joins = [(name, 'arg{}'.format(arg_no+1))
                   for arg_no, name in enumerate(undef_args)]

    wf.connect([
                (infosource,   select_files, [(field, field) for field in undef_args]),
                (select_files, joinpath,     input_joins),
                (joinpath,     datasink,     [("out", "container")]),
               ],
              )

    return wf


